"""quixote.imphooks

Import hooks; when installed, these hooks allow importing .ptl files
as if they were Python modules.

TODO:

    - cache compiled code in .pyc files
    - change priority of .ptl to be higher than .pyc
    - use imputil instead of ihooks?

Note: there's some unpleasant incompatibility between ZODB's import
trickery and the import hooks here.  Bottom line: if you're using ZODB,
import it *before* installing the Quixote/PTL import hooks.
"""

__revision__ = "$Id: ptl_import.py,v 1.21 2001/08/29 21:43:37 nascheme Exp $"


import sys
import os.path
import imp, ihooks, new, string
import cgi, urllib
import struct
import marshal
import stat

from ptl_compile import compile_template, PTL_EXT, PTLC_EXT, PTLC_MAGIC

assert sys.hexversion >= 0x20000b1, "need Python 2.0b1 or later"

def _new_module(code, name, filename):
    new_mod = new.module(name)
    new_mod.__file__ = filename
    new_mod.__name__ = name
    sys.modules[name] = new_mod

    exec code in new_mod.__dict__

    return new_mod

def _load_ptlc(name, filename, file=None):
    if not file:
        try:
            file = open(filename, "rb")
        except IOError:
            return None
    path, ext = os.path.splitext(filename)
    ptl_filename = path + PTL_EXT
    magic = file.read(len(PTLC_MAGIC))
    if magic != PTLC_MAGIC:
        return _load_ptl(name, ptl_filename)
    ptlc_mtime = marshal.load(file)
    try:
        mtime = os.stat(ptl_filename)[stat.ST_MTIME]
    except OSError:
        mtime = ptlc_mtime
    if mtime != ptlc_mtime:
        return _load_ptl(name, ptl_filename)
    code = marshal.load(file)
    return _new_module(code, name, filename)
    
def _load_ptl(name, filename, file=None):
    if not file:
        try:
            file = open(filename, "rb")
        except IOError:
            return None
    path, ext = os.path.splitext(filename)
    ptlc_filename = path + PTLC_EXT
    try:
        output = open(ptlc_filename, "wb")
    except IOError, msg:
        output = None
    try:
        code = compile_template(file, filename, output)
    except:
        # don't leave a corrupt .ptlc file around
        if output:
            output.close()
            os.unlink(ptlc_filename)
        raise
    return _new_module(code, name, filename)


# Constant used to signal a PTL files
PTLC_FILE = 128
PTL_FILE = 129

class PTLHooks (ihooks.Hooks):

    def get_suffixes (self):
        # add our suffixes
        L = imp.get_suffixes()
        return L + [(PTLC_EXT, 'rb', PTLC_FILE), (PTL_EXT, 'r', PTL_FILE)]

class PTLLoader (ihooks.ModuleLoader):
    
    def load_module (self, name, stuff):
        file, filename, info = stuff
        (suff, mode, type) = info

        # If it's a PTL file, load it specially.
        if type == PTLC_FILE:
            return _load_ptlc(name, filename, file)

        elif type == PTL_FILE:
            return _load_ptl(name, filename, file)

        else:
            # Otherwise, use the default handler for loading
            return ihooks.ModuleLoader.load_module( self, name, stuff)


_installed = 0

def install ():
    global _installed
    if not _installed:
        hooks = PTLHooks()
        loader = PTLLoader(hooks)
        ihooks.install(ihooks.ModuleImporter(loader))
        _installed = 1

if __name__ == '__main__':
    import foo
    print repr(foo.f1(1,2,3))



#!/www/python/bin/python

"""
Compile a PTL template.

First the tokens "template" are replaced with "def".  Next, the file is
parsered into an parse tree.  This tree is converted into a modified AST.
It is during this state that the semantics are modified by adding extra
nodes to the tree.  Finally bytecode is generated using the compiler
package.

Note that script/module requires the compiler package.
"""

# created 2000/07/28, nas

__revision__ = "$Id: ptl_compile.py,v 1.17 2001/08/29 17:52:35 gward Exp $"

import sys
import os
import string
import symbol
import token
import parser
import re

from compiler import pycodegen, visitor, transformer, walk
from compiler import ast
from compiler.consts import OP_ASSIGN

# magic names inserted into the code

# try to figure out what package we are in
_package = string.split(__name__, ".")[:-1]
IO_MODULE = string.join(_package + ["TemplateIO"], ".")
IO_CLASS = "TemplateIO"
IO_INSTANCE = "__output"
TEMPLATE_PREFIX = "__template__"

class TemplateTransformer(transformer.Transformer):

    def __init__ (self, *args, **kwargs):
        transformer.Transformer.__init__(self, *args, **kwargs)
        self.__in_template = [] # stack, 1 if in template else 0

    def file_input(self, nodelist):
        # Add a "from IO_MODULE import IO_CLASS" statement to the
        # beginning of the module.
        doc = None # self.get_docstring(nodelist, symbol.file_input)
        if sys.hexversion >= 0x20000b1:
            # imports work differently since Python 2.0b1
            imp = ast.From(IO_MODULE, [(IO_CLASS, None)])
        else:
            imp = ast.From(IO_MODULE, [IO_CLASS])

        # Add an IO_INSTANCE binding for module level expressions (like
        # doc strings).  This instance will not be returned.
        klass = ast.Name(IO_CLASS)
        instance = ast.CallFunc(klass, [])
        assign_name = ast.AssName(IO_INSTANCE, OP_ASSIGN)
        assign = ast.Assign([assign_name], instance)

        stmts = [ imp, assign ]

        for node in nodelist:
            if node[0] != token.ENDMARKER and node[0] != token.NEWLINE:
                self.com_append_stmt(stmts, node)

        return ast.Module(doc, ast.Stmt(stmts))

    def funcdef(self, nodelist):
        lineno = nodelist[1][2]
        name     = nodelist[1][1]
        args     = nodelist[2][2]

        if not name.startswith(TEMPLATE_PREFIX):
            # just a normal function, let base class handle it 
            self.__in_template.append(0)
            n = transformer.Transformer.funcdef(self, nodelist)

        else:
            self.__in_template.append(1)

            name = name[len(TEMPLATE_PREFIX):]

            # Add "IO_INSTANCE = IO_CLASS()" statement at the beginning of
            # the function and a "return IO_INSTANCE" at the end.
            if args[0] == symbol.varargslist:
                names, defaults, flags = self.com_arglist(args[1:])
            else:
                names = defaults = ()
                flags = 0
            doc = None # self.get_docstring(nodelist[4])

            # code for function
            code = self.com_node(nodelist[4])

            # create an instance, assign to IO_INSTANCE
            klass = ast.Name(IO_CLASS)
            instance = ast.CallFunc(klass, [])
            assign_name = ast.AssName(IO_INSTANCE, OP_ASSIGN)
            assign = ast.Assign([assign_name], instance)

            # return the str() of the instance
            func = ast.Getattr(ast.Name(IO_INSTANCE), "__str__")
            ret = ast.Return(ast.CallFunc(func, []))

            # wrap original function code
            code = ast.Stmt([assign, code, ret])

            n = ast.Function(name, names, defaults, flags, doc, code)
            n.lineno = lineno

        self.__in_template.pop()
        return n

    def expr_stmt(self, nodelist):
        if not self.__in_template or not self.__in_template[-1]:
            return transformer.Transformer.expr_stmt(self, nodelist)

        # Instead of discarding objects on the stack, call
        # "IO_INSTANCE.write(obj)".
        exprNode = self.com_node(nodelist[-1])
        if len(nodelist) == 1:
            lval = ast.Name(IO_INSTANCE)
            n = ast.AugAssign(lval, '+=', exprNode)
            if hasattr(exprNode, 'lineno'):
                n.lineno = exprNode.lineno
        elif nodelist[1][0] == token.EQUAL:
            nodes = [ ]
            for i in range(0, len(nodelist) - 2, 2):
                nodes.append(self.com_assign(nodelist[i], OP_ASSIGN))
            n = ast.Assign(nodes, exprNode)
            n.lineno = nodelist[1][2]
        else:
            lval = self.com_augassign(nodelist[0])
            op = self.com_augassign_op(nodelist[1])
            n = ast.AugAssign(lval, op[1], exprNode)
            n.lineno = op[2]
        return n


_template_re = re.compile(r"^([ \t]*)template([ \t]+)", re.MULTILINE)

def translate_tokens(buf, filename):
    """
    Since we can't modify the parser in the builtin parser module we
    must do token translation here.  Luckily it does not affect line
    numbers.
   
    template foo(...): -> def __template__foo(...):
   
    XXX This parser is too stupid.  For example, it doesn't understand
    triple quoted strings.
    """
    global _def_re, _template_re

    # change template to def
    buf = _template_re.sub(r"\1def\2%s" % TEMPLATE_PREFIX, buf)

    # compile() and parsermodule don't accept certain modules if they are
    # missing a trailing newline.  The Python interpreter seems to add a
    # newline when importing modules.  We will match that its behavior.
    if buf and buf[-1] != '\n':
        buf += "\n"

    return buf


def parse(buf, filename='<string>'):
    buf = translate_tokens(buf, filename)
    try:
        return TemplateTransformer().parsesuite(buf)
    except (parser.ParserError, SyntaxError):
        # Let Python compiler generate a better error message
        import __builtin__, sys
        try:
            __builtin__.compile(buf, filename, 'exec')
        except SyntaxError, value:
            # Python does not use filename in the exception but stupidly
            # uses "<string>" instead.  Patch it up ourself.
            lineno = value.lineno
            offset = value.offset
            text = value.text
            raise SyntaxError(str(value), (filename, lineno, offset, text))
        raise # hmm, the Python compiler didn't raise an exception!?


PTL_EXT = ".ptl"
PTLC_EXT = ".ptlc"
if sys.hexversion >= 0x20100b1:
    PTLC_MAGIC = "PTLC\000\005"
elif sys.hexversion >= 0x20000b1:
    PTLC_MAGIC = "PTLC\000\004"
else:
    PTLC_MAGIC = "PTLC\000\003"


class Template(pycodegen.Module):

    def compile(self):
        ast = parse(self.source, self.filename)
        root, filename = os.path.split(self.filename)
        gen = pycodegen.ModuleCodeGenerator(filename)
        walk(ast, gen, 1)
        self.code = gen.getCode()

    def dump(self, f):
        import marshal
        import stat
        import struct

        f.write(PTLC_MAGIC)
        mtime = os.stat(self.filename)[stat.ST_MTIME]
        marshal.dump(mtime, f)
        marshal.dump(self.code, f)


def compile_template(input, filename, output=None):
    """compile_template(input, filename, output=None) -> code
    
    Compile an open file.  If output is not None then the code is written
    with the magic template header.  The code object is returned.
    """
    buf = input.read()
    template = Template(buf, filename)
    template.compile()
    if output:
        template.dump(output)
    return template.code

def compile(inputname, outputname):
    """compile(inputname, outputname)
    
    Compile a template file.  The new template is writen to outputname.
    """
    input = open(inputname)
    output = open(outputname, "wb")
    try:
        compile_template(input, inputname, output)
    except:
        # don't leave a corrupt .ptlc file around
        output.close()
        os.unlink(outputname)
        raise

def main():
    args = sys.argv[1:]
    if not args:
        print "no files to compile"
    else:
        for filename in args:
            path, ext = os.path.splitext(filename)
            compile(filename, path + PTLC_EXT)

if __name__ == "__main__":
    main()



"""quixote

A highly Pythonic web application framework.
"""

__revision__ = "$Id: __init__.py,v 1.9 2001/10/10 18:01:41 gward Exp $"

__version__ = "0.4.1"

__all__ = ['publish', 'compile_template', 'imphooks', 'session', 'errors']

# These are frequently needed by Quixote applications, so make them easy
# to get at.
from quixote.publish import \
     Publisher, get_publisher, get_request, get_session, get_user

# Can't think of anywhere better to put this, so here it is.
def enable_ptl ():
    """
    Installs the import hooks needed to import PTL modules.  This must
    be done explicitly because not all Quixote applications need to use
    PTL, and import hooks are deep magic that can cause all sorts of
    mischief and deeply confuse innocent bystanders.  Thus, we avoid
    invoking them behind the programmer's back.  One known problem is
    that, if you use ZODB, you must import ZODB before calling this
    function.
    """
    from quixote import ptl_import
    ptl_import.install()




_q_exports = ["simple", "error"]

import sys
from quixote.demo.pages import _q_index
from quixote.demo.integer_ui import IntegerUI

def simple (request):
    # This function returns a plain text document, not HTML.
    request.response.set_header("Content-type", "text/plain")
    return "This is the Python function 'quixote.demo.simple'.\n"

def error (request):
    raise ValueError, "this is a Python exception"

def _q_getname(request, component):
    return IntegerUI(request, component)



"""Yet another unit testing framework for Python."""

# created 2000/03/17, Greg Ward

__revision__ = "$Id: unittest.py,v 1.1 2000/10/20 19:37:18 akuchlin Exp $"

import sys, string, re
from types import *
from indented_file import IndentedFile
from traceback import print_exc, print_tb


class TestScenario:

    def __init__ (self, outfile=None):
        self.total_count = 0
        self.total_run = 0
        self.total_ok = 0
        
        self.tests_seen = None
        self.tests_run = 0
        self.tests_ok = 0
        self.test_reports = None
        self.options = None

        if outfile:
            self.out = outfile
        else:
            self.out = IndentedFile (sys.stdout, 0, 2)

    def setup (self):
        raise NotImplementedError, "subclasses must override 'setup()'"

    def shutdown (self):
        raise NotImplementedError, "subclasses must override 'shutdown()'"

    def warn (self, msg):
        sys.stderr.write ("warning: " + msg + "\n")


    def _resolve_case_names (self, case_names):
        """Given a list of test case names, resolves them into (unbound)
        method objects.  This just means look up each string in
        'case_names' in the current instance's class, prepending
        "check_" if needed.  Returns a list of unbound method objects.
        Raises ValueError if any of the names in 'case_names' don't
        resolve to a method in the current class."""
        
        klass = self.__class__
        klass_name = klass.__name__

        methods = []
        for name in case_names:
            if name[0:6] == 'check_':
                try_names = [name]
            else:
                try_names = [name, 'check_' + name]

            for try_name in try_names:
                try:
                    func = getattr (klass, try_name)
                    break
                except AttributeError:
                    continue

            else:                       # all getattr attempts failed
                if len (try_names) == 1:
                    reason = "'%s' method not found" % try_names[0]
                elif len (try_names) == 2:
                    reason = "neither '%s' nor '%s' methods found" % \
                             tuple(try_names)
                else:
                    raise RuntimeError, "this can't happen!"

                raise ValueError, \
                      ("bad test case name '%s': " + reason +
                       " in class %s") % \
                       (name, klass_name)

            methods.append (func)

        return methods

    # _resolve_case_names ()


    def _find_test_cases (self):
        """Examine the current class (ie. the class of 'self') to find
        all test cases in it -- that is, all methods whose names
        begin with 'check_'.  Returns a list of unbound method objects,
        sorted by name."""

        methods = []
        klass = self.__class__

        # Get a list of (name, method) tuples (so we can sort on name)
        for name in vars(klass).keys():
            if name[0:6] == "check_":
                member = getattr (klass, name)
                if type(member) is UnboundMethodType:
                    methods.append ((name, member))
        methods.sort ()

        # Throw out the names -- all we really need are the method
        # objects (unbound!)        
        for i in range(len(methods)):
            methods[i] = methods[i][1]

        return methods

    # _find_test_cases ()
    

    def _determine_test_cases (self, cases=None):
        """Figure out which set of test cases we are supposed to run.
        If 'cases' is supplied, it should be a list (or tuple) of
        strings that name test case methods in the current class (with
        or without the leading 'check_').  Otherwise, if there is an
        attribute 'test_cases' on 'self' (usually a class attribute, but
        you never know), then it is treated the same -- it should be a
        list of method names.  Otherwise, use '_find_test_cases()' to
        find all test cases from the class dictionary."""

        klass = self.__class__
        klass_name = klass.__name__

        if cases is None:

            # Generate 'cases', a list of method names.  Either it comes
            # from the class attribute 'test_cases', or we figure it out
            # ourselves by inspecting the class dictionary.
            if hasattr (self, "test_cases"):
                case_methods = self._resolve_case_names(self.test_cases)
                #cases = self.test_cases
            else:
                case_methods = self._find_test_cases()

        elif type (cases) in (ListType, TupleType):
            case_methods = self._resolve_case_names(cases)

        else:
            raise TypeError, "'cases' must be a list, tuple, or None"

        return case_methods

    # _determine_test_cases ()


    def _get_testcase_info (self, method):

        case_name = method.__name__
        class_name = self.__class__.__name__

        doc = getattr (method, '__doc__')
        if doc is None:
            self.warn ("test case '%s' in class %s: no doc string" %
                       (case_name, class_name))
            msg = case_name
            count = "???"
        else:
            match = re.match (r'(.*):\s*(\d+)$', method.__doc__)
            if match is None:
                self.warn (("test case '%s' in class %s: " +
                            "badly-formatted doc string") %
                           (case_name, class_name))
                msg = doc
                count = "???"
            else:
                (msg, count) = match.group (1, 2)
                count = int (count)

        return (msg, count)

    # _get_testcase_info ()


    def run (self, cases=None, options=None):
        """Run the appropriate test cases in this test scenario.  Test
        cases are specified either by 'cases', a 'test_cases' attribute
        on 'self' (usually a class attribute), or by searching for all
        methods whose names start with 'check_'.  See
        '_determine_test_cases()' for details.

        Every test case is implemented by a method of 'self'.  Those
        methods are called in turn, and the results are reported.  If an
        exception was caught from the body of the test case itself (ie.,
        not from a particular test), that is reported.  The number of
        tests run, passed, and failed is always reported."""

        if options is not None:
            self.options = options
        if self.options is None:
            self.options = TestOptions()

        case_methods = self._determine_test_cases (cases)
        for method in case_methods:
            case_name = method.__name__
            if case_name[0:6] == 'check_':
                abbr_case_name = case_name[6:]
            else:
                abbr_case_name = case_name
            class_name = self.__class__.__name__
            
            (descr, test_count) = self._get_testcase_info (method)
            if test_count == "???" or self.total_count == "???":
                self.total_count = "???"
            else:
                self.total_count = self.total_count + test_count
            if self.options.preface_testcase:
                self.out.write ("%s ('%s'):\n" % (descr, abbr_case_name))

            self.tests_run = 0
            self.tests_ok = 0
            self.test_reports = []
            try:
                self.setup ()
                self.out.indent_more ()
                method (self)
                self.out.indent_less ()
                self.shutdown ()
                testcase_finished = 1
            except:
                self.out.indent_less ()
                exc_info = sys.exc_info()
                exc_name = exception_name(exc_info[0])
                exc_msg = str(exc_info[1])
                self.warn (("test case '%s' in class %s raised\n" +
                            "         %s: %s") %
                           (case_name, class_name, exc_name, exc_msg))
                print_exc (file=self.out)
                testcase_finished = 0
                self.shutdown ()

            # Print a summary of the test case; will be slightly
            # different depending on whether we've already printed the
            # test case description.
            (status, failed, summary) = \
                gen_summary (test_count, self.tests_run, self.tests_ok,
                             testcase_finished,
                             "test", "tests")
            if self.options.show_testcase_summary:
                self.out.write (status + ": ")
                if self.options.preface_testcase:
                    self.out.write (summary + "\n")
                else:
                    self.out.write ("%s ('%s') (%s)\n" %
                                    (descr, abbr_case_name, summary))

            self.total_run = self.total_run + self.tests_run
            self.total_ok = self.total_ok + self.tests_ok

            # If we *didn't* report test results (success/failure) as we
            # ran each test, do so now.
            if not self.options.report_immediate:
                self.out.indent_more()
                for report in self.test_reports:
                    self.print_report(report)
                self.out.indent_less()

            # Carefully consider whether we should print a newline.
            if (method is not case_methods[-1] and
                (self.options.preface_testcase and
                 self.options.show_testcase_summary)
                or
                (failed and
                 not self.options.preface_testcase and
                 self.options.show_testcase_summary)
               ):
                self.out.write ("\n")

        # for case_methods

    # run ()

    def get_stats (self):
        return (self.total_count, self.total_run, self.total_ok)


    # ------------------------------------------------------------------
    # Everything after this point relates to running individual test
    # cases, while everything previous deals with multiple test cases,
    # i.e. the collection of test cases in a particular test scenario.
    # This is a pretty natural dividing line; if I could figure out a
    # good way to divide these into two classes, I would!
    # ------------------------------------------------------------------


    def record (self, status, message, extra=None, traceback=None):
        self.test_reports.append ((status, message, extra, traceback))

    def print_report (self, report=None):
        if report is None:
            report = self.test_reports[-1]
        (status, message, extra, traceback) = report
        self.out.write ("%s: %s\n" % (status, message))
        self.out.indent_more()
        if extra:
            for line in extra:
                self.out.write (line + "\n")
        if traceback and self.options.show_tracebacks:
            print_tb (traceback, file=self.out)
        self.out.indent_less()
            

    def report_pass (self, msg):
        self.tests_ok = self.tests_ok + 1
        if not self.options.report_passes:
            return
        self.record ("ok", msg)
        if self.options.report_immediate:
            self.print_report ()

    def report_fail (self, msg, traceback=None):
        if not self.options.report_failures:
            return
        info = get_caller_info (2)
        failed_at = ["failed at %s, line %s (in '%s()'):" % info[0:3],
                     "  " + info[3]]
        self.record ("not ok", msg, failed_at, traceback)

        if self.options.report_immediate:
            self.print_report ()
        #if self.options.report_immediate:
        #    self.out.write ("not ok: " + msg + "\n")
        #    self.out.indent_more()
        #    self.out.write (failed_at[0] + "\n")
        #    self.out.write (failed_at[1] + "\n")
        #    if traceback and self.options.show_tracebacks:
        #        print_tb (traceback, file=self.out)
        #    self.out.indent_less()


    def test_val (self, code, value=None, match_ident=0, match_types=0):
        (globals, locals) = get_caller_env ()
        self.tests_run = self.tests_run + 1
        try:
            val = eval (code, globals, locals)
        except:
            (exc_seen, exc_value, exc_tb) = sys.exc_info()
            self.report_fail \
                ("%s != %s (raised %s)" %
                 (code, `value`, describe_exc (exc_seen, exc_value)),
                 exc_tb)
        else:
            if match_ident:
                if val is value:
                    self.report_pass ("%s is %s@%x" %
                                      (code, `value`, id(value)))
                else:
                    self.report_fail ("%s is not %s@%x (got %s@%x)" %
                                      (code,
                                       `value`, id(value),
                                       `val`, id(val)))

            else:                       # just match on equality


                if match_types:
                    if type(value) is InstanceType:
                        if isinstance(val, value.__class__):
                            type_msg = (" (and is an instance of %s)" %
                                        value.__class__.__name__)
                            types_ok = 1
                        else:
                            type_msg = (" (type mismatch: not instance of %s)"%
                                        value.__class__.__name__)
                            types_ok = 0
                    elif type(val) is type(value):
                        type_msg = " (and types match: %s)" % type(value)
                        types_ok = 1
                    else:
                        type_msg = (" (type mismatch: expected %s, got %s)" %
                                      (type(value), type(val)))
                        types_ok = 0
                else:
                    type_msg = ""
                    types_ok = 1

                if not types_ok:
                    self.report_fail ("%s != %s" % (code, `value`) +
                                      type_msg)
                else:
                    if val == value:
                        self.report_pass ("%s == %s" % (code, `value`) +
                                          type_msg)
                    else:
                        self.report_fail ("%s != %s (got %s)" %
                                          (code, `value`, `val`))

    # test_val ()

    # Alias for test_val()
    test = test_val

    def test_bool (self, code, want_true=1):
        (globals, locals) = get_caller_env ()
        self.tests_run = self.tests_run + 1
        try:
            val = eval (code, globals, locals)
        except:
            (exc_seen, exc_value, exc_tb) = sys.exc_info()
            self.report_fail ("%s not %s (raised %s)" %
                              (code, want_true and "true" or "false",
                               describe_exc (exc_seen, exc_value)),
                              exc_tb)
        else:
            if want_true and val:
                self.report_pass ("%s is true (%s)" % (code, `val`))
            elif not want_true and not val:
                self.report_pass ("%s is false (%s)" % (code, `val`))
            elif want_true and not val:
                self.report_fail ("%s not true (got %s)" % (code, `val`))
            elif not want_true and val:
                self.report_fail ("%s not false (got %s)" % (code, `val`))
            else:
                raise RuntimeError, "this can't happen!"
                
    # test_bool ()

    def test_seq (self, code, seqval, match_types=0, match_order=1):
        (globals, locals) = get_caller_env ()
        self.tests_run = self.tests_run + 1
        try:
            val = eval (code, globals, locals)
        except:
            (exc_seen, exc_value, exc_tb) = sys.exc_info()
            self.report_fail ("%s != %s (raised %s)" %
                              (code, `seqval`,
                               describe_exc (exc_seen, exc_value)),
                              exc_tb)
        else:

            # If match_order false, then we put the two sequence objects
            # into real lists and sort them before comparing, so that
            # differences in the ordering of elements disappear.
            if not match_order:
                try:
                    val = list (val)
                except TypeError:
                    self.report_fail ("%s not a sequence (got %s)" %
                                      (code, `val`))
                    return

                seqval = list (seqval)
                val.sort()
                seqval.sort()

            # If match_types true, then both the type and elements of the
            # two sequences have to be the same -- ie. we can just use
            # the normal comparison operator.  (Note that "match_types
            # and match_order" is effectively the same as just calling
            # 'test_val()'.)
            if match_types:
                passed = (val == seqval)

            # Otherwise, have to compare element-by-element -- ie. do
            # just what the normal comparison operator does, but don't
            # care about the nature of the sequence objects.  Note that
            # we rely on elements of the two sequences to be equal in
            # the conventional sense -- that is, [1,2,3] and (1,2,3) are
            # equal here, as are [1,2,("foo","bar")] and (1,2,("foo","bar"))
            # -- but NOT [1,2,("foo","bar")] and (1,2,["foo","bar"]).
            else:
                if len (val) == len (seqval):
                    for i in range (len (val)):
                        if val[i] != seqval[i]:
                            passed = 0
                            break
                    else:
                        passed = 1
                else:
                    passed = 0

            # OK, we have now (possibly) reordered the two sequences and
            # then compared them.  'passed' is true if they compared OK.
            ordered = (match_order) and "(in order)" or "(ignoring order)"
            if passed:
                self.report_pass ("%s == %s %s" %
                                  (code, `seqval`, ordered))
            else:
                self.report_fail ("%s != %s %s (got %s)" %
                                  (code, `seqval`, ordered, `val`))


    def test_stmt(self, stmt):
        "Test whether the statement 'stmt' completes without problems."

        (globals, locals) = get_caller_env ()
        self.tests_run = self.tests_run + 1
        try:
            exec stmt in globals, locals
        except:
            # Raised an exception
            (exc_seen, exc_value, exc_tb) = sys.exc_info()
            self.report_fail \
                ("%s raised %s" % 
                 (stmt, 
                  describe_exc (exc_seen, exc_value)),
                 exc_tb)
        else:
            # No exception raised
            self.report_pass ("%s ran quietly" % (stmt))

    # test_stmt ()

    def test_exc(self, stmt, exc):
        """Test whether the statement 'stmt' raises the expected exception."""

        (globals, locals) = get_caller_env ()
        self.tests_run = self.tests_run + 1
        try:
            exec stmt in globals, locals
        except:
            (exc_seen, exc_value, exc_tb) = sys.exc_info()
            if exc_seen is exc:         # good, got the expected exception
                self.report_pass \
                    ("%s raised %s" % (stmt, describe_exc (exc, exc_value)))

            else:                       # exception, but not the expected one
                self.report_fail \
                    ("%s didn't raise %s (raised %s)" % 
                     (stmt, exception_name (exc),
                      describe_exc (exc_seen, exc_value)),
                     exc_tb)
            return exc_value
        else:
            self.report_fail \
                ("%s didn't raise %s" % 
                 (stmt, exception_name (exc)))
            return None

    # test_exc ()
        
# class TestScenario


class TestOptions:
    """Contains the options that control the output and verbosity of
    running a test scenario.  This class exists purely as a holder for
    instance attributes:

      [...]
      report_passes (boolean, default false)
        print a line for each individual test that passes
      report_failures (boolean, default true)
        print a couple of lines for each individual test that fails
      show_tracebacks (boolean, default true)
        print a traceback for each failure that involves an unexpected
        exception
      [...]

    """

    # verbosity levels
    LOW = 0
    MEDIUM = 1                              # the default
    HIGH = 2


    def __init__ (self, verbosity=MEDIUM, show_coverage=0):

        self.show_coverage = show_coverage
        self.show_coverage_lines = 0
        if verbosity == TestOptions.MEDIUM:
            self.preface_testcase = 0
            self.report_immediate = 0
            self.report_passes = 0
            self.report_failures = 1
            self.show_tracebacks = 1
            self.show_testcase_summary = 1
            self.show_scenario_banner = 1
            self.show_scenario_summary = 1
            self.show_script_summary = 1

        elif verbosity == TestOptions.HIGH:
            self.preface_testcase = 1
            self.report_immediate = 1
            self.report_passes = 1
            self.report_failures = 1
            self.show_tracebacks = 1
            self.show_testcase_summary = 1
            self.show_scenario_banner = 1
            self.show_scenario_summary = 1
            self.show_script_summary = 1
            if self.show_coverage:
                self.show_coverage_lines = 1

        elif verbosity == TestOptions.LOW:
            self.preface_testcase = 0
            self.report_immediate = 0
            self.report_passes = 0
            self.report_failures = 1
            self.show_tracebacks = 1
            self.show_testcase_summary = 0
            self.show_scenario_banner = 0
            self.show_scenario_summary = 0
            self.show_script_summary = 1

    # __init__ ()

# class TestOptions


def exception_name (exc):
    """Returns the name of an exception without any package or module
    names cluttering it up.  'exc' must be a string or exception
    class)."""
    if type(exc) is ClassType:
        return exc.__name__
    else:
        return exc


def describe_exc (exc_type, exc_value):
    return "%s: \"%s\"" % (exception_name (exc_type), exc_value)


def gen_summary (num_expected, num_run, num_passed, testcase_finished,
                 what, what_plural):
    if num_expected > 1:
        what = what_plural

    if num_expected == num_run == num_passed and testcase_finished:
        return ("ok", 0, "%d %s passed" % (num_passed, what))
    else:
        # ugh: use %s for num_expected because it can be
        # an integer or "???"
        text = ("%s %s expected, %d run, %d failed" %
                (num_expected, what, num_run, num_run-num_passed))
        if not testcase_finished:
            text = text + ", and test case crashed"

        return ("not ok", 1, text)


def caller (level=1):
    """Return the execution frame for some function up the call stack,
       by default (level=1) the immediate caller.  level=0 gives your
       own execution frame (that of the function that called 'caller');
       level=2 gives that of your caller's caller, and so on.  Raises
       RuntimeError if level is greater than the depth of the call stack."""

    try:
        try:
            jhasdgfjhasdf
        except NameError:
            tb = sys.exc_info()[2]

        # Now 'tb' is the traceback containing the frame for *this* function;
        # at the very least, we want our *caller's* frame, so...
        frame = tb.tb_frame.f_back
        if frame is None:
            raise RuntimeError, "'caller' has no caller!"

        # Now follow 'next' pointers until we've reached the desired level
        while level > 0:
            frame = frame.f_back
            if frame is None:
                raise RuntimeError, 'call stack not deep enough'
            level = level - 1

        return frame

    finally:
        del tb                          # break cycle, prevent leak

# caller ()


def get_caller_env (level=1):
    frame = caller (level+1)
    return (frame.f_globals, frame.f_locals)


def get_frame_info (frame):
    """Extracts and returns a tuple of handy information from an
       execution frame:
         (filename, lineno, function_name, code_line)
    """
    import linecache
    filename = frame.f_code.co_filename
    lineno = frame.f_lineno
    function_name = frame.f_code.co_name
    code_line = string.strip (linecache.getline (filename, lineno))

    return (filename, lineno, function_name, code_line)


def get_caller_info (level=1):
    """Extracts and returns a tuple of handy information from some
       caller's execution frame:
         (filename, lineno, function_name, code_line)
       If level is 1 (the default), gets this information about the
       caller of the function that is calling 'get_caller_info()'."""

    return get_frame_info (caller (level+1))



# eg. [FooTest, BarTest]
# or  [(FooTest, ['check_ding', 'check_dong']), BarTest]

def run_scenarios (scenarios, options=None):
    """Run a bunch of test scenarios.  'scenarios' is a list of test
    scenario descriptions.  A test scenario description is either a
    single class object, in which case all test cases in that test
    scenario are run; or it is a tuple (test_scenario, [case1, case2,
    ...])  where 'test_scenario' is again a class object and 'case1'
    etc.  are the names of the test cases to run.  Test case names are
    the same as the method that implements the test case, with the
    leading "check_" optionally lopped off."""

    # Empty list of test scenarios?  Then get the list of scenarios from
    # the global variable 'test_scenarios' in the main module (test script)
    # or, if that variable isn't found, just find all test scenarios in the
    # main module and run them.
    if not scenarios:
        scenarios = find_scenarios (sys.modules['__main__'])
    
    out = IndentedFile(sys.stdout, 0, 2)

    if options.show_coverage:
        module_dict = vars(sys.modules['__main__'])
        tested_modules = module_dict.get('tested_modules')
        if tested_modules:
            from code_coverage import Coverage
            coverage = Coverage(modules=tested_modules)
            sys.settrace(coverage.trace)

    # List of (count, run, ok) counts: for each test scenario run, the
    # number of tests expected, run, and passed.  At the end we tally
    # these up and return the totals (useful for the "run_tests" script,
    # which un turn tallies up the totals for many test scripts.  Yow!
    scenario_stats = []

    # Now loop over the list of test scenario descriptions.  If we just
    # built our own, then most of this code is unused -- it's only useful
    # if the caller passed in a list of scenario descriptions, and we have
    # to figure out what the nature of the decription is.  Of course, the
    # code at the end -- instantiate and run the test scenario -- is always
    # important!
    for scenario in scenarios:
        if type (scenario) is ClassType:
            klass = scenario
            cases = None

        elif type (scenario) is TupleType and len (scenario) == 2:
            (klass, cases) = scenario

        else:
            raise TypeError, \
                  "every element in 'scenarios' must be either a " + \
                  "class object or 2-tuple: '%s' invalid" % `scenario`

        if not (type(klass) is ClassType and issubclass (klass, TestScenario)):
            raise TypeError, \
                 "test scenario classes must be subclasses of TestScenario: "+\
                 "%s invalid" % klass.__name__


        if options.show_scenario_banner:
            scenario_name = klass.__name__
            if hasattr(klass, 'description'):
                banner = "%s (%s):\n" % (scenario_name, klass.description)
            else:
                banner = "%s:\n" % scenario_name
            out.write(banner)
            out.indent_more()

        scenario_obj = klass (outfile=out)
        scenario_obj.run (cases, options=options)
        out.indent_less()

        counts = scenario_obj.get_stats()
        scenario_stats.append(counts)
        if options.show_scenario_summary:
            out.write("%s: %s\n" % summarize_results(counts))

        if scenario is not scenarios[-1] and options.show_scenario_banner:
            out.write("\n")

    if options.show_coverage:
        if tested_modules is None:
            out.write('Coverage analysis requested but '
                      '"tested_modules" not defined\n')
        else:
            # reload tested modules to get coverage of imports, etc.
            for name in tested_modules:
                module = sys.modules.get(name)
                if module:
                    reload(module)
            sys.settrace(None)
            out.write("code coverage:\n")
            out.indent_more()
            coverage.write_results(out, show_lines=options.show_coverage_lines)
            out.indent_less()

    return sum_test_counts(scenario_stats)

# run_scenarios ()


def sum_test_counts (count_list):
    """sum_test_counts(count_list : [(expected : int,
                                      run : int,
                                      passed : int)])
       -> (total_expected : int, total_run : int, total_passed : int)

    Compute sums of a list of test counts.  'count_list' is a list of
    3-tuples (expected, run, passed): the number of tests expected, run,
    and passed.  Returns a 3-tuple that is the sum of all the tuples in
    'count_list'.
    """
    totals = [0, 0, 0]
    for counts in count_list:
        for i in (0, 1, 2):
            if counts[i] == '???':
                totals[i] = '???'
            elif totals[i] != '???':
                totals[i] = totals[i] + counts[i]
    return tuple(totals)

def tests_successful (counts):
    """Returns true if all tests completed successfully."""
    (expected, run, passed) = counts
    return expected == run == passed

def summarize_results (counts, preface=None):
    """summarize_results(counts : [(int, int, int)],
                         preface : string)
       -> (string, string)

    Return a pair of strings summarizing the test counts statistics in
    'counts', a 3-tuple of (expected, run, passed) test counts.  The
    first string is either "ok" or "not ok"; the second gives more
    detail (possibly including 'preface').  Generally, you can just join
    them together with ":" and print them.
    """
    (expected, run, passed) = counts
    if tests_successful(counts):
        summary = "%d tests passed" % expected
        if preface:
            return ("ok", "%s: %s" % (preface, summary))
        else:
            return ("ok", summary)
    else:
        summary = ("%s tests expected, %d run, %d failed" %
                   (expected, run, run-passed))
        if preface:
            return ("not ok", "%s: %s" % (preface, summary))
        else:
            return ("not ok", summary)


def find_test_scenario (script, name, module=None):

    if module is None:
        module = sys.modules['__main__']
    module_dict = vars(module)

    if name[-4:] != "Test":
        try_names = [name, name + "Test"]
    else:
        try_names = [name]

    for try_name in try_names:
        klass = module_dict.get(try_name)
        if klass:
            break
    else:
        if len (try_names) == 1:
            reason = "'%s' class not found" % try_names[0]
        elif len (try_names) == 2:
            reason = "neither '%s' nor '%s' classes found" % tuple(try_names)
        else:
            raise RuntimeError, "this can't happen!"

        raise ValueError, \
              ("unknown test scenario '%s': " + reason +
               " in script %s") % \
              (name, script)

    if not issubclass (klass, TestScenario):
        raise RuntimeError, \
              ("bad test scenario '%s': " +
               "'%s' not a subclass of TestScenario") % \
              (name, name)

    return klass

# find_test_scenario()
    

def find_scenarios (module):

    """Finds all test scenarios (subclasses of TestScenario) made available
    by module 'module' and returns the list of class objects found.
    "Available" means "listed in the module global variable
    'test_scenarios' if it is defined, or defined in that module if
    not"."""

    mod_dict = vars (module)

    if hasattr (module, 'test_scenarios'):
        scenarios = module.test_scenarios
    else:
        scenarios = []
        for val in mod_dict.values():
            if (type (val) is ClassType and
                issubclass (val, TestScenario) and
                val.__module__ == module.__name__):
                scenarios.append(val)

    return scenarios

# find_scenarios ()


def has_test_case (scenario, casename):
    if hasattr (scenario, casename):
        return casename
    elif casename[0:6] != "check_" and hasattr (scenario, "check_" + casename):
        return "check_" + casename
    else:
        return None


def parse_args (argv=sys.argv):
    """Parse the command-line arguments to a test script.
    Syntax is:
      [-c] [-q|-v] [scenario1][:case1,...] ...
    ie. user may specify any number of test scenarios and any number
    of test cases for each scenario.  Scenarios are specified by the
    name of the class that implements them, and cases by the name
    of the method.  Trailing "Test" is optional for scenario names,
    as is leading "check_" for case names.

    If no scenarios or cases are supplied, all found in the test script
    (__main__ module) are run.  If a scenario is specified without any test
    cases, all cases ("check_" methods or methods listed in class attribute
    'test_cases') are run.  If a test case is specified without any test
    scenario, that test case is run in all test scenarios (this is mainly a
    shortcut for scripts with a single test scenario class).

    Exits with appropriate error message and non-zero exit status if any
    errors found in command-line.  Exits with help message and zero exit
    status if "-h" or "--help" seen on command-line.

    Returns a tuple (scenarios, options); 'options' is a TestOptions
    instance recording the -q/-v/etc. options; scenarios is a list
    of test scenario descriptions as required by 'run_scenarios()'."""

    from getopt import getopt

    script = argv[0]
    args = argv[1:]
    usage = "usage: %s: [-q] [-v] [scenario1][:case1,...] ..." % script

    # Check for help options
    if "-h" in argv or "--help" in argv:
        print usage
        sys.exit()

    # Look for -q and -v; modify default option settings appropriately
    (opts, args) = getopt (args, 'cqv')
    verbosity = TestOptions.MEDIUM
    show_coverage = 0
    for (opt,_) in opts:
        if opt == '-c':
            show_coverage = 1
        elif opt == '-q':
            verbosity = TestOptions.LOW
        elif opt == '-v':
            verbosity = TestOptions.HIGH
    options = TestOptions (verbosity, show_coverage)

    all_scenarios = find_scenarios (sys.modules['__main__'])

    # No arguments -- we'll just run all tests
    if not args:
        return (all_scenarios, options)

    scenarios = []
    for arg in args:

        colon = string.find (arg, ':')

        # No colon in 'arg' -- it's just a bare word.
        if colon == -1:

            # Special case for lazy typists running test scripts that only
            # define one test scenario: a lone word on the command line is
            # first checked to see if it's the name of a test case (method)
            # in that test scenario.
            if len(all_scenarios) == 1:
                casename = has_test_case (all_scenarios[0], arg)
                if casename:
                    scenarios.append ((all_scenarios[0],[casename]))
                    continue

            # Otherwise, assume the lone word is the name of a test
            # scenario class; dig up and validate the class object.
            klass = find_test_scenario (script, arg)
            scenarios.append (klass)

        # Found a colon: split into scenario name and list of test cases.
        else:
            scenario = arg[:colon]
            klass = find_test_scenario (script, scenario)

            # If no cases were listed (ie. there's nothing after the
            # colon), then treat this like the no-colon case above: will
            # run all test cases
            if colon+1 == len(arg):
                scenarios.append (klass)

            # Otherwise, pull out the list of case names after the colon
            else:
                cases = string.split (arg[colon+1:], ',')
                scenarios.append ((klass, cases))

        # if/else colon
        
    # for args

    return (scenarios, options)

# parse_args ()


def run_scenario (test_class):
    """Run all test cases in a single test scenario.  Deprecated."""
    scenario = test_class ()
    if not isinstance (scenario, TestScenario):
        raise TypeError, "'test_class' must be a subclass of TestScenario"

    scenario.run ()



#!/usr/bin/env python

"""Run all test scripts in a directory tree and report the results.
Searches for any file "test_*.py" in any subdirectory "test" under the
current directory (or a starting point given on the command line),
and runs them all.
"""

# created 2000/07/10, Greg Ward

__revision__ = "$Id: run_tests.py,v 1.1 2000/10/20 19:36:41 akuchlin Exp $"

import sys, os, re
import getopt
import new
from quixote.test.unittest import TestOptions, \
     find_scenarios, run_scenarios, \
     sum_test_counts, summarize_results, tests_successful

USAGE = """\
usage: %(prog)s [ -q ] -r [start_dir ...]
   or: %(prog)s [ -q ] test_script ...
"""

VERBOSE = 1

def write_now(msg):
    sys.stdout.write(msg)
    sys.stdout.flush()


def find_test_scripts (start_dir):

    """Recursively explore from 'start_dir', looking for all
    "test/test_*.py" files.  Return the list of matches.
    """

    filename_re = re.compile('^test_\w+\.py$')

    def visit (matches, dirname, names, filename_re=filename_re):
        if os.path.basename(dirname) == 'test':
            for name in names:
                if filename_re.match(name):
                    name = os.path.normpath(os.path.join(dirname, name))
                    matches.append(name)

    matches = []
    os.path.walk(start_dir, visit, matches)
    return matches


def run_test_script (script):
    """Run all the tests in a test script.
    """

    # XXX should catch exceptions from execfile here, but that would
    # mean duplicating logic from TestScenario.run(); I'll wait 'till I
    # have a decent "test collection" abstraction and just do it once.

    test_name = os.path.splitext(os.path.basename(script))[0]
    test_mod = new.module(test_name)
    test_mod.__file__ = script + " (fake)"
    execfile(script, test_mod.__dict__)
    scenarios = find_scenarios(test_mod)

    options = TestOptions(verbosity=TestOptions.LOW, show_coverage=0)
    stats = run_scenarios(scenarios, options)
    if VERBOSE or not tests_successful(stats):
        print "%s: %s" % summarize_results(stats, script)
    return stats


def run_all_scripts (scripts):
    script_stats = []                   # list of test count 3-tuples
    for script in scripts:
        script_stats.append(run_test_script(script))

    totals = sum_test_counts(script_stats)
    print "%s: %s" % summarize_results(totals)
    return tests_successful(totals)
    


def main (prog, args):
    global VERBOSE

    usage = USAGE % vars()
    try:
        (opts, args) = getopt.getopt(args, 'rq')
    except getopt.error, msg:
        raise SystemExit, usage + "\n" + msg
    recursive = ('-r', '') in opts
    VERBOSE = ('-q', '') not in opts

    if recursive:
        start_dirs = args or ["."]
        scripts = []
        if VERBOSE:
            write_now("looking for test scripts...")
        for start_dir in start_dirs:
            scripts.extend(find_test_scripts(start_dir))
        if VERBOSE:
            write_now("found %d\n" % len(scripts))
    else:
        if len(args) == 0:
            raise SystemExit, usage + "\nNot enough arguments"
        scripts = args

    success = run_all_scripts(scripts)
    if not success:
        sys.exit(2) # some tests failed

if __name__ == "__main__":
    main(os.path.basename(sys.argv[0]), sys.argv[1:])



"""quixote.http_response

Provides the HTTPResponse class.

Derived from Zope's HTTPResponse module.
"""

##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################

__revision__ = "$Id: http_response.py,v 1.25 2001/09/19 14:51:32 nascheme Exp $"

import sys
from types import StringType, IntType

status_reasons = {
    100: 'Continue',
    101: 'Switching Protocols',
    102: 'Processing',
    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non-Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',
    207: 'Multi-Status',
    300: 'Multiple Choices',
    301: 'Moved Permanently',
    302: 'Moved Temporarily',
    303: 'See Other',
    304: 'Not Modified',
    305: 'Use Proxy',
    307: 'Temporary Redirect',
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Time-out',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Request Entity Too Large',
    414: 'Request-URI Too Large',
    415: 'Unsupported Media Type',
    416: 'Requested range not satisfiable',
    417: 'Expectation Failed',
    422: 'Unprocessable Entity',
    423: 'Locked',
    424: 'Failed Dependency',
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Time-out',
    505: 'HTTP Version not supported',
    507: 'Insufficient Storage',
}


class HTTPResponse:
    """
    An object representation of an HTTP response.
    
    The Response type encapsulates all possible responses to HTTP
    requests.  Responses are normally created by the Quixote publisher
    or by the HTTPRequest class (every request must have a response,
    after all).

    Instance attributes:
      status_code : int
        HTTP response status code (integer between 100 and 599)
      reason_phrase : string
        the reason phrase that accompanies status_code (usually
        set automatically by the set_status() method)
      headers : { string : string }
        most of the headers included with the response; every header set
        by 'set_header()' goes here.  Does not include "Status" or
        "Set-Cookie" headers (unless someone uses set_header() to set
        them, but that would be foolish).
      body : string
        the response body, None by default.  If the body is never
        set (ie. left as None), the response will not include 
        "Content-type" or "Content-length" headers.  These headers
        are set as soon as the body is set (with set_body()), even
        if the body is an empty string.
      cookies : { name:string : { attrname : value } }
        collection of cookies to set in this response; it is expected
        that the user-agent will remember the cookies and send them on
        future requests.  The cookie value is stored as the "value"
        attribute.  The other attributes are as specified by RFC 2109.
      cache : boolean = false
        if false, instruct the user-agent not to cache the response
        body.  (Currently done with an HTTP 1.0-style "Pragma: no-cache"
        header.)

    """

    def __init__ (self, status=200, body=None):
        """
        Creates a new HTTP response.
        """
        self.set_status(status)
        self.headers = {}
        
        if body is not None:
            self.set_body(body)
        else:
            self.body = None

        self.cookies = {}
        self.cache = 0

    def set_status (self, status, reason=None):
        """set_status(status : int, reason : string = None)

        Sets the HTTP status code of the response.  'status' must be an
        integer in the range 100 .. 599.  'reason' must be a string; if
        not supplied, the default reason phrase for 'status' will be
        used.  If 'status' is a non-standard status code, the generic
        reason phrase for its group of status codes will be used; eg.
        if status == 493, the reason for status 400 will be used.
        """
        if type(status) is not IntType:
            raise TypeError, "status must be an integer"
        if not (100 <= status <= 599):
            raise ValueError, "status must be between 100 and 599"

        self.status_code = status
        if reason is None:
            if status_reasons.has_key(status):
                reason = status_reasons[status]
            else:
                # Eg. for generic 4xx failures, use the reason
                # associated with status 400.
                reason = status_reasons[(status / 100) * 100]
        else:
            reason = str(reason)

        self.reason_phrase = reason

    def set_header (self, name, value, literal=0):
        """set_header(name : string, value : string)

        Sets an HTTP return header "name" with value "value", clearing
        the previous value set for the header, if one exists.
        """
        self.headers[name.lower()] = value

    def set_body (self, body):
        """set_body(body : string)
        
        Sets the return body equal to the (string) argument "body". Also
        updates the "Content-length" return header.  Sets the
        "Content-type" header if it hasn't already been set.
        """
        if type(body) is not StringType:
            raise TypeError("'body' must be a string")

        self.body = body
        self.set_header('content-length', len(body))
        if not self.headers.has_key('content-type'):
            self.set_header('content-type', 'text/html; charset=iso-8859-1')

    def expire_cookie (self, name, **attrs):
        """
        Cause an HTTP cookie to be removed from the browser
        
        The response will include an HTTP header that will remove the cookie
        corresponding to "name" on the client, if one exists.  This is
        accomplished by sending a new cookie with an expiration date
        that has already passed.  Note that some clients require a path
        to be specified - this path must exactly match the path given
        when creating the cookie.  The path can be specified as a keyword
        argument.
        """
        dict = {'max_age': 0, 'expires': 'Thu, 01-Jan-1970 00:00:00 GMT'}
        dict.update(attrs)
        self.set_cookie(name, "deleted", **dict)

    def set_cookie (self, name, value, **attrs):
        """set_cookie(name : string, value : string, **attrs)

        Set an HTTP cookie on the browser.

        The response will include an HTTP header that sets a cookie on
        cookie-enabled browsers with a key "name" and value "value".
        Cookie attributes such as "expires" and "domains" may be
        supplied as keyword arguments; see RFC 2109 for a full list.
        (For the "secure" attribute, use any true value.)

        This overrides any previous value for this cookie.  Any
        previously-set attributes for the cookie are preserved, unless
        they are explicitly overridden with keyword arguments to this
        call.
        """
        cookies = self.cookies
        if cookies.has_key(name):
            cookie = cookies[name]
        else:
            cookie = cookies[name] = {}
        cookie.update(attrs)
        cookie['value'] = value

    def redirect (self, location, permanent=0):
        """Cause a redirection without raising an error"""
        if not isinstance(location, StringType):
            raise TypeError, "location must be a string (got %s)" % `location`
        # Ensure that location is a full URL
        if location.find('://') == -1:
            raise ValueError, "URL must include the server name"
        if permanent:
            status = 301
        else:
            status = 302
        self.set_status(status)
        self.headers['location'] = location

        doc = """\
<html>
 <head><title>Redirect to %(location)s</title></head>
 <body>
 <h1>Redirect</h1>
 <p>Your browser should have redirected you to
 <center><a href="%(location)s">%(location)s</a></center>
 </p>
 </body>
</html>
""" % vars()

        return doc


    def _gen_cookie_headers (self):
        """_gen_cookie_headers() -> [string]

        Build a list of "Set-Cookie" headers based on all cookies
        set with 'set_cookie()', and return that list.
        """
        cookie_list = []
        for (name, attrs) in self.cookies.items():

            # Note that as of May 98, IE4 ignores cookies with
            # quoted cookie attr values, so only the value part
            # of name=value pairs may be quoted.

            # 'chunks' is a list of "name=val" chunks; will be joined
            # with "; " to create the "Set-cookie" header.
            chunks = ['%s="%s"' % (name, attrs['value'])] 
                
            for (name, val) in attrs.items():
                name = name.lower()
                if name in ('expires', 'domain', 'path', 'max_age', 'comment'):
                    name = name.replace('_', '-')
                    chunks.append("%s=%s" % (name, val))
                elif name == 'secure' and v:
                    chunks.append("secure")

            cookie = "Set-Cookie: " + ("; ".join(chunks))
            cookie_list.append(cookie)

        # Should really check size of cookies here!
        
        return cookie_list

    def write (self, file):
        """write(file : file)

        Write the HTTP response headers and body to 'file'.  This is not
        a complete HTTP response, as it doesn't start with a response
        status line as specified by RFC 2616.  It does, however, start
        with a "Status" header as described by the CGI spec.  It
        is expected that this response is parsed by the web server
        and turned into a complete HTTP response.
        """

        # XXX currently we write a response like this:
        #  Status: 200 OK
        #  Content-type: text/html; charset=iso-8859-1
        #  Content-length: 100
        #  Set-Cookie: foo=bar
        #  Set-Cookie: bar=baz
        #
        #  <html><body>This is a document</body></html>
        #
        # which has to be interpreted by the web server to create
        # a true HTTP response -- that is, this is for a
        # "parsed header" CGI driver script.
        #
        # We should probably have provisions for operating in
        # "non-parsed header" mode, where the CGI script is responsible
        # for generating a complete HTTP response with no help from the
        # server.

        headers = self.headers

        # "Status" header must come first.
        file.write("Status: %03d %s\r\n" % (self.status_code,
                                            self.reason_phrase))

        # Now the bulk of the headers (anything set with 'set_header()',
        # most likely "Content-type" and "Content-length").
        for (key, val) in headers.items():
            file.write("%s: %s\r\n" % (key, val))

        # All the "Set-Cookie" headers.
        if self.cookies:
            for hdr in self._gen_cookie_headers():
                file.write(hdr + "\r\n")

        # Cache directives (HTTP 1.0 only, arggh).
        if not self.cache:
            file.write("Pragma: no-cache\r\n")

        # And the respone body.
        file.write("\r\n")
        if self.body is not None:
            file.write(self.body)



"""quixote.errors

Exception classes used by Quixote
"""

class PublishError (Exception):
    """PublishError exceptions are raised due to some problem with the
    data provided by the client and are raised during the publishing
    process.  Quixote will abort the current request and return an error
    to the client.  The error returned to the client is controlled by the
    format method.

    The formatting done by the Quixote versions of these exceptions is
    very simple.  Applications will probably wish to raise application
    specific subclasses which do more sophisticated formatting.
    """

    def format (self, request):
        """format(request : HTTPRequest) -> string

        Called by Quixote to return a response to the user.
        """
        return None                     # use default formatting


class TraversalError (PublishError):
    """
    Raised when a client attempts to access a resource that does not
    exist or is otherwise unavailable to them (eg. a Python function
    not listed in its module's _q_exports list).

    public_msg should be a user-readable message that reveals no
    inner workings of your application; it will always be shown.  If
    not supplied, an innocuous default ("Resource not found") will be
    shown.

    private_msg will only be shown if the config option
    SECURE_ERRORS is false; Quixote uses this to give you more
    detail about why a resource was not found -- eg. Python module
    couldn't be imported, name exists but not in _q_exports, etc.
    You might want to use it for similar, application-specific
    information.  (SECURE_ERRORS should always be true in a
    production environment, since these details about the inner
    workings of your application could conceivably be useful to
    attackers.)

    path should be the path to the requested resource; if not
    supplied, the current request object will be fetched and its
    get_path() method called.
    """

    # XXX it should be possible to make publish raise an application
    # specific subclass of this exception

    def __init__ (self, public_msg = "Resource not found",
                  private_msg=None, path=None):
        if private_msg is None: private_msg = public_msg
        if path is None:
            import quixote
            path = quixote.get_request().get_path()

        self.public_msg = public_msg
        self.private_msg = private_msg
        self.path = path

    def __str__ (self):
        if self.private_msg:
            return "%s: %s" % (self.path, self.private_msg)
        else:
            return self.path


class QueryError (PublishError):
    """Should be raised if bad data was provided in the query part of a
    URL or in the content of a POST request.  What constitutes bad data is
    solely application dependent (eg: letters in a form field when the
    application expects a number).

    The formatting done by this exception is very simple.  Applications
    will probably wish to raise subclasses which do more sophisticated
    formatting.
    """

    def __init__ (self, msg=None):
        self.msg = msg or "Invalid query"

    def __str__ (self):
        return self.msg


class AccessError (PublishError):
    """Should be raised raised if the client does not have access to the
    requested resource.  Usually applications will raise this error from
    an _q_access method.
    """

    def __init__ (self, msg=None):
        self.msg = msg or "Access denied"
    
    def __str__ (self):
        return self.msg


class SessionError (PublishError):
    """Raised when a session cookie has an invalid session ID.  This
    could be either a broken/malicious client or an expired session.
    """

    def __init__ (self, msg=None, session_id=None):
        self.msg = msg or "Expired or invalid session ID"
        self.session_id = session_id

    def __str__ (self):
        if self.session_id is None:
            return self.msg
        else:
            return "%s: %s" % (self.msg, self.session_id)



__revision__ = "$Id: TemplateIO.py,v 1.6 2001/08/23 19:08:44 nascheme Exp $"

class TemplateIO:
    """Collect output for PTL scripts.
    
    Behaves similar to the StringIO class except that str() is called on
    input and None objects are not added to the output.
    """

    def __init__(self):
        self.data = []

    def __iadd__(self, other):
        if other is not None:
            self.data.append(other)
        return self

    def __add__(self, other):
        result = TemplateIO()
        result.data = self.data[:]
        result.data.append(other)
        return result
            
    def __radd__(self, other):
        result = TemplateIO()
        result.data = [other]
        result.data.extend(self.data)
        return result
        
    def __str__(self):
        return ''.join(map(str, self.data))

    def __repr__(self):
        return ("<%s at %x: %d chunks>" %
                (self.__class__.__name__, id(self), len(self.data)))



"""quixote.mod_python_handler

mod_python handler for Quixote.  See the "mod_python configuration"
section of doc/web-server.txt for details.
"""

__revision__ = "$Id: mod_python_handler.py,v 1.5 2001/09/28 21:03:22 gward Exp $"

# adapted from Erno Kuusela's quixote_handler.py, 2001/09/19 GPW

import os, sys
from mod_python import apache
from quixote import Publisher, enable_ptl

class ModPythonPublisher(Publisher):
    def publish_modpython(self, req):
        """publish_modpython() -> None
        
        Entry point from mod_python.
        """
        real_env = os.environ
        try:
            os.environ = apache.build_cgi_env(req)
            self.publish(apache.CGIStdin(req),
                         apache.CGIStdout(req),
                         sys.stderr)
        finally:
            os.environ = real_env
        return apache.OK

enable_ptl()

name2publisher = {}

def handler(req):
    opts = req.get_options()
    try:
        package = opts['quixote-root-namespace']
    except KeyError:
        package = None

    if not package:
        return apache.HTTP_INTERNAL_SERVER_ERROR
    else:
        if not name2publisher.has_key(package):
            name2publisher[package] = ModPythonPublisher(package)
        return name2publisher[package].publish_modpython(req)



"""quixote.form.form

Provides the Form class and bureaucracy for registering widget classes.
(The standard widget classes are registered automatically.)
"""

# created 2000/09/19, GPW

__revision__ = "$Id: form.py,v 1.7 2001/09/19 15:23:36 gward Exp $"

from types import StringType
from quixote import get_session
from quixote.form import form_templates
from quixote.form.widget import FormValueError, Widget, HiddenWidget


class FormTokenWidget (HiddenWidget):
    def parse (self, request):
        # keep self.value rather than setting it from request.form
        return request.form.get(self.name)


class Form:
    """
    A form is the major element of an interactive web page.  A form
    consists of the following:
      * widgets (input/interaction elements)
      * text
      * layout
      * code to process the form

    All four of these are the responsibility of Form classes.
    Typically, you will create one Form subclass for each form in your
    application.  Thanks to the separation of responsibilities here,
    it's not too hard to structure things so that a given form is
    rendered and/or processed somewhat differently depending on context.
    That separation is as follows:
      * the constructor declares what widgets are in the form, and
        any static text that is always associated with those widgets
        (in particular, a widget title and "hint" text)
      * the 'render()' method combines the widgets and their associated
        text to create a (1-D) stream of HTML that represents the
        (2-D) web page that will be presented to the user
      * the 'process()' method parses the user input values from the form
        and validates them
      * the 'action()' method takes care of finishing whatever action
        was requested by the user submitting the form -- commit
        a database transaction, update session flags, redirect the
        user to a new page, etc.

    This class provides a default 'process()' method that just parses
    each widget, storing any error messages for display on the next
    'render()', and returns the results (if the form parses
    successfully) in a dictionary.

    This class also provides a default 'render()' method that lays out
    widgets and text in a 3-column table: the first column is the widget
    title, the second column is the widget itself, and the third column is
    any hint and/or error text associated with the widget.  Also provided
    are methods that can be used to construct this table a row at a time,
    so you can use this layout for most widgets, but escape from it for
    oddities.

    Instance attributes:
      widgets : { widget_name:string : widget:Widget }
        dictionary of all widgets in the form
      widget_order : [Widget]
        same widgets as 'widgets', but ordered (because order matters)
      submit_buttons : [SubmitButtonWidget]
        the submit button widgets in the form

      error : { widget_name:string : error_message:string }
      hint : { widget_name:string : hint_text:string }
      title : { widget_name:string : widget_title:string }
      required : { widget_name:string : boolean }

    """

    TOKEN_NAME = "_form_id" # name of hidden token widget

    def __init__ (self):

        # The first major component of a form: its widgets.  We want
        # both easy access and order, so we have a dictionary and a list
        # of the same objects.  The dictionary is keyed on widget name.
        # These are populated by the 'add_*_widget()' methods.
        self.widgets = {}
        self.widget_order = []
        self.submit_buttons = []
        self.cancel_url = None

        # The second major component: text.  It's up to the 'render()'
        # method to figure out how to lay these out; the standard
        # 'render()' does so in a fairly sensible way that should work
        # for most of our forms.  These are also populated by the
        # 'add_*_widget()' methods.
        self.error = {}
        self.hint = {}
        self.title = {}
        self.required = {}

        # unique token for each form, this prevents many cross-site attacks
        # and prevents a form from being submitted twice
        self.add_widget(FormTokenWidget, self.TOKEN_NAME,
                        value=get_session().create_form_token())

        # Subclasses should override this method to specify the actual
        # widgets in this form -- typically this consists of a series of
        # calls to 'add_widget()', which updates the data structures we
        # just defined.


    # The third major component: layout.  This combines text and widgets
    # in a 1-D stream of HTML, or in a 2-D web page (depending on your
    # level of abstraction).

    def render (self, request, action_url):
        """render(request : HTTPRequest,
                  action_url : string)
           -> HTML text

        Render a form as HTML.  Return an object whose 'str()'
        representation is the HTML code for the form.
        """
        return form_templates.render(self, request, action_url)

    # Other rendering methods: these are needed by the default
    # 'render()', and will also be useful for forms that implement a
    # custom 'render()'.
    _render_start = form_templates._render_start
    _render_finish = form_templates._render_finish
    _render_sep = form_templates._render_sep
    _render_widget_row = form_templates._render_widget_row
    _render_error = form_templates._render_error
    _render_hint = form_templates._render_hint
    _render_hidden_widgets = form_templates._render_hidden_widgets
    _render_submit_buttons = form_templates._render_submit_buttons


    # The fourth and final major component: code to process the form.
    # The standard 'process()' method just parses every widget and
    # returns a { field_name : field_value } dictionary as 'data'.

    def process (self, request):
        """process(request : HTTPRequest) -> values : { string : any }

        Process the form data, validating all input fields (widgets).
        If any errors in input fields, adds error messages to the
        'error' attribute (so that future renderings of the form will
        include the errors).  Returns a dictionary mapping widget names to
        parsed values.
        """
        self.error.clear()

        values = {}
        for widget in self.widget_order:
            try:
                val = widget.parse(request)
            except FormValueError, exc:
                self.error[widget.name] = str(exc)
            else:
                values[widget.name] = val

        return values

    def action (self, request, submit, values):
        """action(request : HTTPRequest, submit : string,
                  values : { string : any }) -> string

        Carry out the action required by a form submission.  'submit' is the
        name of submit button used to submit the form.  'values' is the
        dictionary of parsed values from 'process()'.  Note that error
        checking cannot be done here.  It must done in the 'process()'
        method
        """
        raise NotImplementedError, "sub-classes must implement 'action()'"
    
    def handle (self, request):
        """handle(request : HTTPRequest) -> string 

        Master method for handling forms.  It should be called after
        initializing a form.  Controls form action based on a request.  You
        probably should override 'process' and 'action' instead of
        overriding this method.
        """
        action_url = request.get_path()
        query = request.get_environ("QUERY_STRING")
        if query:
            action_url += "?" + query
        if not self.form_submitted(request):
            return self.render(request, action_url)
        submit = self.get_submit_button(request)
        if submit == "cancel":
            return request.redirect(self.cancel_url)
        values = self.process(request)
        if submit == "":
            # The form was submitted by unknown submit button, assume that
            # the submission was required to update the layout of the form.  
            # Clear the errors and re-render the form.
            self.error.clear()
            return self.render(request, action_url)

        # before calling action() ensure that there is a valid token present
        token = values.get(self.TOKEN_NAME)
        if not request.session.has_form_token(token):
            if not self.error:
                # if there are other errors then don't show the token error,
                # the form needs to be resubmitted anyhow
                self.error[self.TOKEN_NAME] = (
                       "The form you have submitted is invalid.  It has "
                       "already been submitted or has expired. Please "
                       "resubmit the form.")
        else:
            request.session.remove_form_token(token)

        if self.error:
            return self.render(request, action_url)
        else:
            return self.action(request, submit, values)
            

    # -- Convenience methods -------------------------------------------

    def form_submitted (self, request):
        """form_submitted(request : HTTPRequest) -> boolean
        
        Return true if a form was submitted in the current request.
        """
        return len(request.form) > 0

    def get_submit_button (self, request):
        """get_submit_button(request : HTTPRequest) -> string | None

        Get the name of the submit button that was used to submit the
        current form.  If the browser didn't include this information in
        the request, use the first submit button registered.
        """
        for button in self.submit_buttons:
            if request.form.has_key(button.name):
                return button.name
        else:
            if request.form and self.submit_buttons:
                return ""
            else:
                return None

    def get_widget (self, widget_name):
        return self.widgets.get(widget_name)

    def parse_widget (self, name, request):
        """parse_widget(name : string, request : HTTPRequest) -> any

        Parse the value of named widget.  If any parse errors, store the
        error message (in self.error) for use in the next rendering of
        the form and return None; otherwise, return the value parsed
        from the widget (whose type depends on the widget type).
        """
        try:
            return self.widgets[name].parse(request)
        except FormValueError, exc:
            self.error[name] = str(exc)
            return None

    def store_value (self, widget_name, request, target,
                     mode="modifier",
                     key=None,
                     missing_error=None):
        """store_value(widget_name : string,
                       request : HTTPRequest,
                       target : instance | dict,
                       mode : string = "modifier",
                       key : string = widget_name,
                       missing_error : string = None)

        Parse a widget and, if it parsed successfully, store its value
        in 'target'.  The value is stored in 'target' by name 'key';
        if 'key' is not supplied, it defaults to 'widget_name'.
        How the value is stored depends on 'mode':
          * modifier: call a modifier method, eg. if 'key' is "foo",
            call 'target.set_foo(value)'
          * direct: direct attribute update, eg. if 'key' is
            "foo" do "target.foo = value"
          * dict: dictionary update, eg. if 'key' is "foo" do
            "target['foo'] = value"

        If 'missing_error' is supplied, use it as an error message if
        the field doesn't have a value -- ie. supplying 'missing_error'
        means this field is required.
        """
        value = self.parse_widget(widget_name, request)
        if (value is None or value == "") and missing_error:
            self.error[widget_name] = missing_error
            return None

        if key is None:
            key = widget_name
        if mode == "modifier":
            # eg. turn "name" into "target.set_name", and
            # call it like "target.set_name(value)"
            mod = getattr(target, "set_" + key)
            mod(value)
        elif mode == "direct":
            if not hasattr(target, key):
                raise AttributeError, \
                      ("target object %s doesn't have attribute %s" % 
                       (`target`, key))
            setattr(target, key, value)
        elif mode == "dict":
            target[key] = value
        else:
            raise ValueError, "unknown update mode %s" % `mode`

    def clear_widget (self, widget_name):
        self.widgets[widget_name].clear()

    def get_widget_value (self, widget_name):
        return self.widgets[widget_name].value

    def set_widget_value (self, widget_name, value):
        self.widgets[widget_name].set_value(value)

    # -- Form population methods ---------------------------------------

    def add_widget (self, widget_type, name, value=None,
                    title=None, hint=None, required=0, **args):
        """add_widget(widget_type : string | Widget,
                      name : string,
                      value : any = None,
                      title : string = None,
                      hint : string = None,
                      required : boolean = 0,
                      ...)

        Create a new Widget object and add it to the form.  The widget
        class used depends on 'widget_type', and the expected type of
        'value' also depends on the widget class.  Any extra keyword
        args are passed to the widget constructor.
        """
        if self.widgets.has_key(name):
            raise ValueError, "form already has '%s' variable" % name
        klass = get_widget_class(widget_type)
        new_widget = apply(klass, (name, value), args)

        self.widgets[name] = new_widget
        self.widget_order.append(new_widget)
        self.title[name] = title
        self.hint[name] = hint
        self.required[name] = required

    def add_submit_button (self, name, value):
        global _widget_class
        if self.widgets.has_key(name):
            raise ValueError, "form already has '%s' variable" % name
        new_widget = _widget_class['submit_button'](name, value)

        self.widgets[name] = new_widget
        self.submit_buttons.append(new_widget)

    def add_cancel_button (self, caption, url):
        if not isinstance(url, StringType):
            raise TypeError, "url must be a string (got %r)" % url
        self.add_submit_button("cancel", caption)
        self.cancel_url = url

# class Form


_widget_class = {}

def register_widget_class (klass, widget_type=None):
    global _widget_class
    if widget_type is None:
        widget_type = klass.widget_type
    assert widget_type is not None, "widget_type must be defined"
    _widget_class[widget_type] = klass

def get_widget_class (widget_type):
    global _widget_class
    return _widget_class.get(widget_type, widget_type)



"""mems.tools.code_coverage

Provides code coverage analysis for testing
"""

# created 2000/07/06, Neil Schemenauer based on trace.py

__revision__ = "$Id: code_coverage.py,v 1.1 2000/10/20 20:56:22 akuchlin Exp $"

# Example Usage:
#
# import code_coverage
# co = code_coverage.Coverage()
# co.run('main()')
# co.write_results()
#
# or
#
# $ python code_coverage.py "import mymodule; mymodule.main()"

# trace.py
# Copyright 2000, Mojam Media, Inc., all rights reserved.
# Author: Skip Montanaro
#
# Copyright 1999, Bioreason, Inc., all rights reserved.
# Author: Andrew Dalke
#
# Copyright 1995-1997, Automatrix, Inc., all rights reserved.
# Author: Skip Montanaro
#
# Copyright 1991-1995, Stichting Mathematisch Centrum, all rights reserved.
#
#
# Permission to use, copy, modify, and distribute this Python software and
# its associated documentation for any purpose without fee is hereby
# granted, provided that the above copyright notice appears in all copies,
# and that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of neither Automatrix,
# Bioreason or Mojam Media be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior permission.

import sys, os, string

# Given a code string, return the SET_LINENO information
def _find_LINENO_from_string(co_code):
    """return all of the SET_LINENO information from a code string"""
    import dis
    linenos = {}

    # This code was filched from the `dis' module then modified
    n = len(co_code)
    i = 0
    prev_op = None
    prev_lineno = 0
    while i < n:
	c = co_code[i]
	op = ord(c)
	if op == dis.SET_LINENO:
	    if prev_op == op:
		# two SET_LINENO in a row, so the previous didn't
		# indicate anything.  This occurs with triple
		# quoted strings (?).  Remove the old one.
		del linenos[prev_lineno]
	    prev_lineno = ord(co_code[i+1]) + ord(co_code[i+2])*256
	    linenos[prev_lineno] = 1
	if op >= dis.HAVE_ARGUMENT:
	    i = i + 3
	else:
	    i = i + 1
	prev_op = op
    return linenos

def _find_LINENO(code):
    """return all of the SET_LINENO information from a code object"""
    import types

    # get all of the lineno information from the code of this scope level
    linenos = _find_LINENO_from_string(code.co_code)

    # and check the constants for references to other code objects
    for c in code.co_consts:
	if type(c) == types.CodeType:
	    # find another code object, so recurse into it
	    linenos.update(_find_LINENO(c))
    return linenos

def find_executable_linenos(filename):
    """return a dict of the line numbers from executable statements in a file

    Works by finding all of the code-like objects in the module then searching
    the byte code for 'SET_LINENO' terms (so this won't work one -O files).

    """
    import parser

    prog = open(filename).read()
    ast = parser.suite(prog)
    code = parser.compileast(ast, filename)

    # The only way I know to find line numbers is to look for the
    # SET_LINENO instructions.  Isn't there some way to get it from
    # the AST?
    
    return _find_LINENO(code)

class Ignore:
    def __init__(self, modules = None, dirs = None):
	self._mods = modules or []
	self._dirs = dirs or []

	self._ignore = { '<string>': 1 }


    def names(self, filename, modulename):
	if self._ignore.has_key(modulename):
	    return self._ignore[modulename]

	# haven't seen this one before, so see if the module name is
	# on the ignore list.  Need to take some care since ignoring
	# "cmp" musn't mean ignoring "cmpcache" but ignoring
	# "Spam" must also mean ignoring "Spam.Eggs".
	for mod in self._mods:
	    if mod == modulename:  # Identical names, so ignore
		self._ignore[modulename] = 1
		return 1
	    # check if the module is a proper submodule of something on
	    # the ignore list
	    n = len(mod)
	    # (will not overflow since if the first n characters are the
	    # same and the name has not already occured, then the size
	    # of "name" is greater than that of "mod")
	    if mod == modulename[:n] and modulename[n] == '.':
		self._ignore[modulename] = 1
		return 1

	# Now check that __file__ isn't in one of the directories
	if filename is None:
	    # must be a built-in, so we must ignore
	    self._ignore[modulename] = 1
	    return 1

	# Ignore a file when it contains one of the ignorable paths
	for d in self._dirs:
	    # The '+ os.sep' is to ensure that d is a parent directory,
	    # as compared to cases like:
	    #  d = "/usr/local"
	    #  filename = "/usr/local.py"
	    # or
	    #  d = "/usr/local.py"
	    #  filename = "/usr/local.py"
	    if string.find(filename, d + os.sep) == 0:
		self._ignore[modulename] = 1
		return 1

	# Tried the different ways, so we don't ignore this module
	self._ignore[modulename] = 0
	return 0

class Coverage:
    def __init__(self, modules=None):
        """modules : list of modules for which to show coverage"
        default is all modules"""
        self.modules = {} # module -> filename
	self.counts = {}  # (modulename, linenumber) -> count
        if modules is None:
            self.show_all = 1
        else:
            self.show_all = 0
            for module in modules:
                self.modules[module] = None

    def trace(self, frame, why, arg):
	if why == 'line':
            modulename = frame.f_globals["__name__"]
            if self.show_all or self.modules.has_key(modulename):
                # something is fishy about getting the file name
                filename = frame.f_globals.get("__file__", None)
                if filename is None:
                    filename = frame.f_code.co_filename
		self.modules[modulename] = filename

		lineno = frame.f_lineno

		# record the file name and line number of every trace
		key = (modulename, lineno)
		self.counts[key] = self.counts.get(key, 0) + 1

	return self.trace

    def run(self, cmd):
        import __main__
        dict = __main__.__dict__
        sys.settrace(self.trace)
        try:
            exec cmd in dict, dict
        finally:
            sys.settrace(None)

    def runctx(self, cmd):
        import __main__
        dict = __main__.__dict__
        sys.settrace(self.trace)
        try:
            exec cmd in dict, dict
        finally:
            sys.settrace(None)

    def runfunc(self, func, *args, **kw):
        result = None
        sys.settrace(self.trace)
        try:
            result = apply(func, args, kw)
        finally:
            sys.settrace(None)
        return result

    def write_results(self, outfile=sys.stdout, show_lines=1):
        import re
        # turn the counts data ("(module, lineno) = count") into something
        # accessible on a per-module basis
        per_module = {}
        for modulename, lineno in self.counts.keys():
            try:
                lines_hit = per_module[modulename]
            except KeyError:
                lines_hit = per_module[modulename] = {}
            lines_hit[lineno] = self.counts[(modulename, lineno)]

        # there are many places where this is insufficient, like a blank
        # line embedded in a multiline string.
        blank = re.compile(r'^\s*(#.*)?$')

        for modulename, filename in self.modules.items():
            if filename == '<string>':
                continue
            outfile.write('%s: ' % modulename)
            if filename is None:
                # module as not executed as all
                outfile.write("not executed\n")
                continue
            if filename[-4:] == ".pyc" or filename[-4:] == ".pyo":
                orig_filename = filename[:-4] + '.py'
            else:
                orig_filename = filename

            if show_lines:
                outfile.write('\n')

            # Get the original lines from the .py file
            try:
                lines = open(orig_filename, 'r').readlines()
            except IOError, err:
                sys.stderr.write(
                    "%s: Could not open %s for reading because: %s - skipping\n" % \
                    ("trace", `filename`, err.strerror))
                continue

            # get a list of the line numbers which represent executable
            # content (returned as a dict for better lookup speed)
            executable_linenos = find_executable_linenos(orig_filename)

            lines_hit = per_module[modulename]
            num_lines = 0
            num_lines_hit = 0
            for i in range(len(lines)):
                line = lines[i]

                # do the blank/comment match to try to mark more lines
                # (help the reader find stuff that hasn't been covered)
                if lines_hit.has_key(i+1):
                    num_lines_hit = num_lines_hit + 1
                    num_lines = num_lines + 1
                    if show_lines:
                        # count precedes the lines that we captured
                        outfile.write('%5d: ' % lines_hit[i+1])
                elif blank.match(line):
                    # blank lines and comments are preceded by dots
                    if show_lines:
                        outfile.write('    . ')
                else:
                    # lines preceded by no marks weren't hit
                    # Highlight them if so indicated, unless the line contains
                    # '#pragma: NO COVER' (it is possible to embed this into
                    # the text as a non-comment; no easy fix)
                    if executable_linenos.has_key(i+1) and \
                       string.find(lines[i], '#pragma NO COVER') == -1:
                        num_lines = num_lines + 1
                        if show_lines:
                            outfile.write('>>>>>> ')
                    else:
                        if show_lines:
                            outfile.write(' '*7)
                if show_lines:
                    outfile.write(string.expandtabs(lines[i], 8))
            if num_lines == 0:
                num_lines = 1 # avoid blowup on empty files
            outfile.write("%3.1f%% (%d/%d)\n" %
                    ((100.0*num_lines_hit)/num_lines, num_lines_hit, num_lines))



def main(stmt):
    co = Coverage()
    co.runctx(stmt)
    co.write_results()
    

if __name__ == '__main__':
    import sys
    if len(sys.argv) == 2:
        main(sys.argv[1])
    else:
        print "Usage: code_coverage.py <statement>"
        sys.exit(1)



"""Various functions for dealing with HTML.

These functions are fairly simple but it is critical that they be used
correctly.  Many security problems are caused by quoting errors (cross site
scripting is one example).


html_quote
----------

Use for quoting data that will occur outside a HTML tag, for example:

    "<title>%s</title>" % html_quote(title)


value_quote
-----------

Use for quoting HTML attribute values, for example:

    '<input type=text value=%s>' % value_quote(name)


url_quote
---------

Use for quoting data to be included as part of a URL, for example:

    '<a href="/search?keyword=%s">' % url_quote(input)

Note that since href is an attribute, its value should usually be quoted
with value_quote.  To see why this is necessary, consider the URL:

    http://example.com/?a=1&copy=0

As a URL this is perfectly valid.  However, when placed in an HTML document
the string "&copy" should be interpreted as the copyright character.

"""

# created 20010821, NAS (from mems.ui.lib.util and mems.ui.lib.html)

__revision__ = "$Id: html.py,v 1.2 2001/09/28 19:22:21 nascheme Exp $"

import cgi
import urllib
from types import UnicodeType


def _quote (value, quoter, fallback):
    if value is None:
        if fallback is None:
            raise ValueError, "value is None and no fallback supplied"
        else:
            return fallback
    if isinstance(value,  UnicodeType):
        value = value.encode('iso-8859-1')
    return quoter(str(value))
 

def html_quote(value, fallback=None):
    """html_quote(value : any [, fallback : string]) -> string

    Quotes 'value' for use in an HTML page; see cgi.escape().  If value is
    None, then the behavior depends on the fallback argument.  If it is not
    supplied then an error is raised.  Otherwise, the fallback value is
    returned unquoted.
    """
    return _quote(value, cgi.escape, fallback)


def url_quote(value, fallback=None):
    """url_quote(value : any [, fallback : string]) -> string

    Quotes 'value' for use in a URL; see urllib.quote().  If value is None,
    then the behavior depends on the fallback argument.  If it is not
    supplied then an error is raised.  Otherwise, the fallback value is
    returned unquoted.
    """
    return _quote(value, urllib.quote, fallback)


def value_quote(value):
    """Quote HTML attribute values.
    
    XHTML 1.0 requires that all values be quoted.  weblint claims
    that some clients don't understand single quotes.  For compatibility
    with HTML, XHTML 1.0 requires that ampersands be encoded.
    """
    assert value is not None, "can't pass None to value_quote"
    value = str(value).replace('&', '&amp;')
    value = value.replace('"', '&quot;')
    return '"%s"' % value


def link (url, text, title=None, name=None):
    return render_tag("a", name=name, href=url, title=title) + str(text) + "</a>"


ValuelessAttr = ["valueless_attr"] # magic singleton object

def render_tag (tag, **attrs):
    r = ""
    r += "<%s" % tag
    for (attr, val) in attrs.items():
        if val is ValuelessAttr:
            r += " %s" % attr
        elif val is not None:
            r += " %s=%s" % (attr, value_quote(val))
    r += ">"
    return r



"""quixote.form

The web interface framework, consisting of Form and Widget base classes
(and a bunch of standard widget classes recognized by Form).
Application developers will typically create a Form subclass for each
form in their application; each form object will contain a number
of widget objects.  Custom widgets can be created by inheriting
and/or composing the standard widget classes.
"""

# created 2000/09/19 - 22, GPW

__revision__ = "$Id: __init__.py,v 1.1 2001/08/22 14:37:10 nascheme Exp $"

from quixote.form.form import Form, register_widget_class
from quixote.form.widget import Widget, StringWidget, PasswordWidget, \
    TextWidget, CheckboxWidget, RadiobuttonsWidget, SingleSelectWidget, \
    SelectWidget, OptionSelectWidget, MultipleSelectWidget, ListWidget, \
    SubmitButtonWidget, HiddenWidget, FloatWidget, IntWidget, FormValueError

# Register the standard widget classes
register_widget_class(StringWidget)
register_widget_class(PasswordWidget)
register_widget_class(TextWidget)
register_widget_class(CheckboxWidget)
register_widget_class(RadiobuttonsWidget)
register_widget_class(SingleSelectWidget)
register_widget_class(OptionSelectWidget)
register_widget_class(MultipleSelectWidget)
register_widget_class(ListWidget)
register_widget_class(SubmitButtonWidget)
register_widget_class(HiddenWidget)
register_widget_class(FloatWidget)
register_widget_class(IntWidget)



"""quixote.form.widget

Provides the basic web widget classes: Widget itself, plus StringWidget,
TextWidget, CheckboxWidget, etc.
"""

# created 2000/09/20, GPW

__revision__ = "$Id: widget.py,v 1.5 2001/10/09 21:07:10 abenard Exp $"

from types import *
from quixote import get_request
from quixote.html import html_quote, render_tag, ValuelessAttr


class FormValueError (Exception):
    """Raised whenever a widget has problems parsing its value."""


class Widget:
    """Abstract base class for web widgets.  The key elements
    of a web widget are:
      - name
      - data type (what the widget returns)
      - widget type (how the widget looks/works in the browser)
      - value

    The name and value are instance attributes (because they're specific to
    a particular widget in a particular context); the data type and widget
    type are class attributes.

    Instance attributes:
      name : string
      value : any        # depends on data_type class attribute

    Feel free to access these directly; to set them, use the 'set_*()'
    modifier methods.
    """

    # Subclasses must define these.  'widget_type' is just a string,
    # e.g.  "string", "text", "checkbox".  'data_type' is a string
    # describing the (Python) type of data returned by the widget, eg.
    # "string", "int", "boolean", "[string]", "(string, number)".
    widget_type = None
    data_type = None

    def __init__ (self, name, value=None):
        assert self.__class__ is not Widget, "abstract class"
        self.name = name
        self.value = value

    def __repr__ (self):
        return "<%s at %x: %s>" % (self.__class__.__name__,
                                   id(self),
                                   self.name)

    def __str__ (self):
        return "%s: %s" % (self.widget_type, self.name)

    def set_name (self, name):
        self.name = name

    def set_value (self, value):
        self.value = value

    def clear (self):
        self.value = None


    # -- Subclasses must implement these -------------------------------

    def render (self, request):
        """render(request) -> HTML text"""
        raise NotImplementedError

    def parse (self, request):
        """parse(request) -> any (depends on data_type)"""
        value = request.form.get(self.name)
        if type(value) is StringType and value.strip():
            self.value = value
        else:
            self.value = None

        return self.value


    # -- Convenience methods for subclasses ----------------------------

    # This one's really only for composite widgets; lives here until
    # we have a demonstrated need for a CompositeWidget class.
    def get_subwidget_name (self, name):
        return "%s.%s" % (self.name, name)

    def create_subwidget (self, widget_type, widget_name, value=None, **args):
        from quixote.form.form import get_widget_class
        klass = get_widget_class(widget_type)
        name = self.get_subwidget_name(widget_name)
        return apply(klass, (name, value), args)


# class Widget


# -- Fundamental widget types ------------------------------------------
# These correspond to the standard types of input tag in HTML:
#   text     StringWidget
#   password PasswordWidget
#   radio    RadiobuttonWidget
#   checkbox CheckboxWidget
# 
# and also to the other basic form elements:
#   <textarea>  TextWidget
#   <select>    SingleSelectWidget
#   <select multiple>
#               MultipleSelectWidget


class StringWidget (Widget):
    """Widget for entering a single string: corresponds to
    "<input type=text>" in HTML.

    Instance attributes:
      value : string
      size : int
      maxlength : int
    """

    widget_type = "string"
    data_type = "string"

    # This lets PasswordWidget be a trivial subclass
    html_type = "text"

    def __init__ (self, name, value=None,
                  size=None, maxlength=None):
        Widget.__init__(self, name, value)
        self.size = size
        self.maxlength = maxlength


    def render (self, request):
        return render_tag("input",
                          type=self.html_type,
                          name=self.name,
                          size=self.size,
                          maxlength=self.maxlength,
                          value=self.value)


class PasswordWidget (StringWidget):
    """Trivial subclass of StringWidget for entering passwords (different
    widget type because HTML does it that way).

    Instance attributes: none
    """

    widget_type = "password"
    html_type = "password"


class TextWidget (Widget):
    """Widget for entering a long, multi-line string; corresponds to
    the HTML "<textarea>" tag.

    Instance attributes:
      value : string
      cols : int
      rows : int
      wrap : string
        (see an HTML book for details on text widget wrap options)
    """

    widget_type = "text"
    data_type = "string"

    def __init__ (self, name, value=None,
                  cols=None, rows=None, wrap='physical'):
        Widget.__init__(self, name, value)
        self.cols = cols
        self.rows = rows
        self.wrap = wrap

    def render (self, request):
        return (render_tag("textarea", name=self.name,
                           cols=self.cols,
                           rows=self.rows,
                           wrap=self.wrap) +
                html_quote(self.value, "") +
                "</textarea>")


    def parse (self, request):
        value = Widget.parse(self, request)
        if value:
            value = value.replace("\r\n", "\n")
            self.value = value
        return self.value


class CheckboxWidget (Widget):
    """Widget for a single checkbox: corresponds to "<input
    type=checkbox>".  Do not put multiple CheckboxWidgets with the same
    name in the same form.

    Instance attributes:
      value : boolean
    """

    widget_type = "checkbox"
    data_type = "boolean"

    def __init__ (self, name, value=None):
        Widget.__init__(self, name, value)

    def render (self, request):
        return render_tag("input",
                          type="checkbox",
                          name=self.name,
                          value="yes",
                          checked=self.value and ValuelessAttr or None)

    def parse (self, request):
        self.value = request.form.has_key(self.name)
        return self.value


class SelectWidget (Widget):
    """Widget for single or multiple selection; corresponds to
    <select name=...>
      <option value=foo>Foo
      ...
    </select>

    Instance attributes:
      allowed_values : [any]
      descriptions : [string]
      quote : boolean = true
      size : int
    """

    # NB. 'data_type' and 'widget_type' not set here because this is
    # an abstract class: they're set by subclasses SingleSelectWidget
    # and MultipleSelectWidget.

    def __init__ (self, name, value=None,
                  allowed_values=None,
                  descriptions=None,
                  quote=1,
                  size=None):
        assert self.__class__ is not SelectWidget, "abstract class"
        Widget.__init__(self, name, value)
        self.quote = quote
        self.set_allowed_values(allowed_values, descriptions)
        self.size = size


    def set_allowed_values (self, allowed_values, descriptions):
        assert type(allowed_values) in (ListType, TupleType), (
            "allowed_values for '%s' not a list: got %s" % (self.name,
                                                            `allowed_values`))
        self.allowed_values = allowed_values
        if descriptions is None:
            self.descriptions = []
            for v in self.allowed_values:
                if v is None:
                    v = ""
                else:   
                    v = str(v)
                self.descriptions.append(v)
        else:
            assert type(descriptions) in (ListType, TupleType), (
                "descriptions for '%s' not a list: got %s" % (self.name,
                                                              `descriptions`))
            assert len(self.allowed_values) == len(descriptions), (
                "allowed_values and descriptions must be the same length: "
                "not %s and %s" % (len(self.allowed_values),
                                   len(descriptions)))
            self.descriptions = descriptions
        if self.quote:
            self.descriptions = map(html_quote, self.descriptions)


    def is_selected (self, value):
        return value == self.value


    def render (self, request):
        if self.widget_type == "multiple_select":
            multiple = ValuelessAttr
        else:
            multiple = None
        if self.widget_type == "option_select":
            onchange = "submit()"
        else:
            onchange = None
        tags = [render_tag("select", name=self.name,
                           multiple=multiple, onchange=onchange,
                           size=self.size)]
        
        for i in range(len(self.allowed_values)):
            if (self.value is not None and
                self.is_selected(self.allowed_values[i])):
                selected = ValuelessAttr
            else:
                selected = None
            r = render_tag("option",
                           value=str(i),
                           selected=selected)
            tags.append(str(r) + self.descriptions[i])
        tags.append("</select>")
        return "\n".join(tags)


class SingleSelectWidget (SelectWidget):
    """Widget for single selection.

    Instance attributes:
      value : any
    """

    widget_type = "single_select"
    data_type = "any"

    def parse (self, request):
        value = request.form.get(self.name)
        self.value = None
        if value:
            if type(value) is ListType:
                raise FormValueError, "cannot select multiple values"
            try:
                index = int(value)
            except ValueError:
                pass
            else:
                if 0 <= index < len(self.allowed_values):
                    self.value = self.allowed_values[index]
        return self.value


class RadiobuttonsWidget (SingleSelectWidget):
    """Widget for a *set* of related radiobuttons -- all have the
    same name, but different values (and only one of those values
    is returned by the whole group).

    Instance attributes:
      value : any
      allowed_values : [any] or (any,)
        the list of possible radiobutton values; in the absence of bugs
        or mischievous clients, 'value' will be one of these.
      descriptions : [string]
        the user-visible radiobutton values: should correspond one-to-one
        to the elements of 'allowed_values', or be None or empty (in
        which case the user-visible values will be the same as the
        behind-the-scenes "real" values)
      quote : boolean = true
        if true, descriptions are HTML-quoted at render time.  If
        your descriptions are known to be HTML text already, set
        this to false
      delim : string = None
        string to emit between each radiobutton in the group.  If
        None, a single newline is emitted.
    """

    widget_type = "radiobuttons"
    data_type = "any"

    def __init__ (self, name, value=None,
                  allowed_values=None,
                  descriptions=None,
                  quote=1,
                  delim=None):
        Widget.__init__(self, name, value)
        self.quote = quote
        self.set_allowed_values(allowed_values, descriptions)
        self.delim = delim or "\n"


    def render (self, request):
        tags = []
        for i in range(len(self.allowed_values)):
            if self.value is not None and (
                self.allowed_values[i] is self.value or # XXX Not sure if this
                self.allowed_values[i] == self.value):  # is the right check
                checked = ValuelessAttr
            else:
                checked = None
            r = render_tag("input",
                           type="radio",
                           name=self.name,
                           value=str(i),
                           checked=checked)
            tags.append(str(r) + self.descriptions[i])
        return self.delim.join(tags)


class MultipleSelectWidget (SelectWidget):
    """Widget for multiple selection.

    Instance attributes:
      value : [any]
    """
        
    widget_type = "multiple_select"
    data_type = "[any]"

    def is_selected (self, value):
        if type(self.value) in (ListType, TupleType) and value in self.value:
            return 1
        return value == self.value


    def append_value (self, value):
        try:
            index = int(value)
        except ValueError:
            pass
        else:
            if 0 <= index < len(self.allowed_values):
                self.value.append(self.allowed_values[index])


    def parse (self, request):
        value = request.form.get(self.name)
        self.value = []
        if value:
            if type(value) is ListType:
                for val in value:
                    self.append_value(val)
            else:
                self.append_value(value)
        if not self.value:
            self.value = None

        return self.value


class SubmitButtonWidget (Widget):
    """
    Instance attributes:
      value : boolean
    """

    widget_type = "submit_button"
    data_type = "boolean"               # ??? whatever...

    def render (self, request):
        assert self.value is not None, "submit button must have a value"
        return render_tag("input",
                          type="submit",
                          name=self.name,
                          value=html_quote(self.value))


    def parse (self, request):
        return request.form.get(self.name)


class HiddenWidget (Widget):
    """
    Instance attributes:
      value : string
    """

    widget_type = "hidden"
    data_type = "string"

    def render (self, request):
        if self.value is None:
            value = None
        else:
            value = html_quote(self.value)
        return render_tag("input",
                          type="hidden",
                          name=self.name,
                          value=value)


# -- Derived widget types ----------------------------------------------
# (these don't correspond to fundamental widget types in HTML,
# so they're separated)

class NumberWidget (StringWidget):
    """
    Instance attributes: none
    """

    # Parameterize the number type (either float or int) through
    # these class attributes:
    type_object = None                  # eg. IntType, FloatType
    type_name = None                    # eg. "int", "float"
    type_converter = None               # eg int(), float() (builtin function)
    type_error = None                   # human-readable error message

    def __init__ (self, name,
                  value=None,
                  size=None, maxlength=None):
        assert self.__class__ is not NumberWidget, "abstract class"
        if value is not None:
            if type(value) is not self.type_object:
                raise TypeError, ("'value' not %s: %s" %
                                  (self.type_name, `value`) )
            value = str(value)
        StringWidget.__init__(self, name, value, size, maxlength)

    def parse (self, request):
        value = StringWidget.parse(self, request)
        if value:
            try:
                self.value = self.type_converter(value)
            except ValueError:
                raise FormValueError, self.type_error
        return self.value


class FloatWidget (NumberWidget):
    """
    Instance attributes:
      value : int
    """

    widget_type = "float"
    data_type = "float"
    type_object = FloatType
    type_name = "float"
    type_converter = float
    type_error = "must be a number"


class IntWidget (NumberWidget):
    """
    Instance attributes:
      value : float
    """

    widget_type = "int"
    data_type = "int"
    type_object = IntType
    type_name = "int"
    type_converter = int
    type_error = "must be an integer"


class OptionSelectWidget (SingleSelectWidget):
    """Widget for single selection with automatic submission.

    Instance attributes:
      value : any
    """

    widget_type = "option_select"
    data_type = "any"

    def __init__(self, *args, **kwargs):
        if kwargs.has_key("options"):
            kwargs["allowed_values"] = kwargs["options"]
            del kwargs["options"]
        SingleSelectWidget.__init__(self, *args, **kwargs)

        request = get_request()
        if request.form:
            SingleSelectWidget.parse(self, request)
        if self.value is None:
            self.value = self.allowed_values[0]

    def render (self, request):
        return (SingleSelectWidget.render(self, request) +
                '<noscript>'
                '<input type=submit name="" value=apply>'
                '</noscript>')

    def parse (self, request):
        return self.value

    def get_current_option (self):
        return self.value


class ListWidget (Widget):
    """Widget for lists of objects.

    Instance attributes:
      value : [any]
    """

    widget_type = "list"
    data_type = "[any]"

    def __init__ (self, name, value=None,
                  element_type=None,
                  element_name="row",
                  **args):
        assert value is None or type(value) is ListType, (
            "form value '%s' not a ListType: got %s" % (name,
                                                        `value`))
        assert type(element_name) is StringType, (
            "form value '%s' element_name not a string: "
            "got %s" % (name, `element_name`))
        
        Widget.__init__(self, name, value)

        if element_type is None:
            self.element_type = "string"
        else:
            self.element_type = element_type
        self.args = args

        self.added_elements_widget = self.create_subwidget(
            "hidden", "added_elements",
            value="1")
        self.add_button = self.create_subwidget(
            "submit_button", "add_element",
            value="Add %s" % element_name)
        
        self.element_widgets = []
        self.element_count = 0

        if self.value is not None:
            for element in self.value:
                self.add_element(element)

        request = get_request()
        
        self.added_elements = int(self.added_elements_widget.parse(request) or
                                  1)

        for i in range(self.added_elements):
            self.add_element()


    def add_element (self, value=None):
        self.element_widgets.append(
            self.create_subwidget(self.element_type,
                                  "element_%d" % self.element_count,
                                  value=value,
                                  **self.args))
        self.element_count += 1


    def render (self, request):
        tags = []
        for element_widget in self.element_widgets:
            tags.append(str(element_widget.render(request)))
        tags.append(self.add_button.render(request))
        tags.append(self.added_elements_widget.render(request))
        return '<br>\n'.join(tags)
        

    def parse (self, request):
        if self.add_button.parse(request):
            self.add_element()
            self.added_elements += 1
            self.added_elements_widget.value = str(self.added_elements)
        self.value = []
        for element_widget in self.element_widgets:
            value = element_widget.parse(request)
            if value is not None:
                self.value.append(value)
        self.value = self.value or None
        return self.value





#! /usr/bin/env python
"""
Dump the information contained in a compiled PTL file.  Based on the
dumppyc.py script in the Tools/compiler directory of the Python
distribution.
"""

__revision__ = "$Id: ptlc_dump.py,v 1.2 2001/08/29 17:53:22 gward Exp $"

import marshal
import dis
import types

from compile_template import PTLC_MAGIC

def dump(obj):
    print obj
    for attr in dir(obj):
        print "\t", attr, repr(getattr(obj, attr))

def loadCode(path):
    f = open(path)
    magic = f.read(len(PTLC_MAGIC))
    if magic != PTLC_MAGIC:
        raise ValueError, 'bad .ptlc magic for file "%s"' % path
    mtime = marshal.load(f)
    co = marshal.load(f)
    f.close()
    return co

def walk(co, match=None):
    if match is None or co.co_name == match:
        dump(co)
        print
        dis.dis(co)
    for obj in co.co_consts:
        if type(obj) == types.CodeType:
            walk(obj, match)

def main(filename, codename=None):
    co = loadCode(filename)
    walk(co, codename)

if __name__ == "__main__":
    import sys
    if len(sys.argv) == 3:
        filename, codename = sys.argv[1:]
    else:
        filename = sys.argv[1]
        codename = None
    main(filename, codename)



"""quixote.session

Quixote session management.  There are three levels to Quixote's
session management system:
  - SessionManager
  - Session
  - ApplicationState

A SessionManager is responsible for creating sessions, setting and reading
session cookies, maintaining the collection of all sessions, and so forth.
There is one SessionManager instance per Quixote process.

A Session is the umbrella object for a single session (notionally, a (user,
host, browser_process) triple).  Simple applications can probably get away
with putting all session data into a Session object (or, better, into an
application-specific subclass of Session).

Larger and more sophisticated applications, though, will want to create
ApplicationState subclasses.  These let you partition your session data
according to the major subdivisions of your web site.  Every Session
instances maintains a collection of ApplicationState instances -- actually,
instances of ApplicationState subclasses, since ApplicationState should not
be instantiated directly.  Each ApplicationState subclass will have at most
one instance in a given Session.

The default implementation provided here is not persistent: when the
Quixote process shuts down, all session data is lost.  You will need to
subclass SessionManager if you want to implement persistent sessions.
"""

__revision__ = "$Id: session.py,v 1.38 2001/09/06 16:10:33 nascheme Exp $"

import sys, string
from time import time, localtime, strftime, strptime, mktime, clock

from quixote import get_publisher
from quixote.errors import SessionError

# This is used both to generate cookies and parse them (when we need
# to figure out how old they are).
TIME_FORMAT = "%Y%m%d:%H%M%S"


def packbytes(s):
    "convert a string of bytes into a long integer"
    n = 0L
    for b in s:
        n <<= 8
        n |= ord(b)
    return n
    
try:
    # /dev/urandom is just as good as /dev/random for cookies (assuming
    # SHA-1 is secure) and it never blocks.
    open("/dev/urandom")
    def randlong(bytes):
        """Return bits of random data as a long integer."""
        return packbytes(open("/dev/urandom").read(bytes))

except IOError:
    # this is much less secure than the above function
    import sha
    _randstate = sha.new(str(time() + clock()))
    def randlong(bytes):
        """Return bits of random data as a long integer."""
        global _randstate
        s = ""
        while len(s) < bytes:
            _randstate.update(str(time() + clock()))
            s += _randstate.digest()
        return packbytes(s[:bytes])
        

class SessionManager:
    """
    SessionManager acts as a dictionary of all sessions, mapping
    session ID strings to individual session objects.  Session objects
    are instances of Session (or an application-specific subclass).
    SessionManager is also responsible for creating and destroying
    sessions, for generating and interpreting session cookies, and for
    session persistence (if any -- this implementation is not
    persistent).

    Most applications can just use this class directly; sessions will
    be kept in memory-based dictionaries, and will be lost when the
    Quixote process dies.  Alternatively an application can subclass
    SessionManager to implement specific behaviour, such as persistence.

    Instance attributes:
      sessions : { session_id:string : Session }
        the collection of sessions managed by this SessionManager

    """

    def __init__ (self):
        self.sessions = {}

    # Methods to be overridden by subclasses
    
    def abort(self):
        "Abort all changes to sessions"
        pass

    def commit(self):
        "Commit all changes to sessions"
        pass

    def new_session(self, request, id):
        "Return a new session object."
        # Subclasses can override this to return an instance of a different
        # class.
        return Session(request, id)

    
    # -- Dictionary methods --------------------------------------------

    def keys(self):
        return self.sessions.keys()

    def values(self):
        return self.sessions.values()

    def items(self):
        return self.sessions.items()

    def get(self, key):
        return self.sessions.get(key)

    def __getitem__(self, key):
        return self.sessions[key]

    def __delitem__ (self, key):
        del self.sessions[key]

    # and here's one real dictionaries *should* have ;-)
    def sorted_keys (self):
        k = self.sessions.keys()
        return k


    # -- Worker methods ------------------------------------------------
    # This is where the heavy lifting of session management
    # gets done.

    def get_session (self, request):
        """get_session(request : HTTPRequest) -> Session

        Fetch or create a session object for the current session, and
        return it.  If a session cookie is found in the HTTP request
        object 'request', uses it to look up and return an existing
        session object.  If no session cookie is found, or its value is
        bogus (ie. not an existing session), a new session is created.
        """
        config = get_publisher().config
        id = request.cookies.get(config.cookie_name)
        if id == "" or id == "*del*":
            id = None
        if id is not None:
            session = self.sessions.get(id)
            if session is None:
                # Note that it's important to revoke the session cookie
                # so the user doesn't keep getting "Expired session ID"
                # error pages.  However, it has to be done in the
                # response object for the error document, which doesn't
                # exist yet.  Thus it's handled by
                # Publisher.format_session_error().
                raise SessionError(session_id=id)
            
        if id is None or session is None:

            # Generate a "session ID", which is just the value of the
            # session cookie we are about to drop on the user
            while id is None or self.sessions.has_key(id):
                # IP address + date/time + big (64-bit) random number
                id = ("%s:%s:%016X" %
                      (request.get_environ("REMOTE_ADDR"),
                       strftime(TIME_FORMAT, localtime(time())),
                       randlong(8)))
                    
            # If we get here, it means this is the first we have seen of
            # this user in this session (or they've been rejecting our
            # cookies, which means we'll keep coming back here).  So, we
            # create a session object -- typically one per (user, host,
            # browser_process) triple
            session = self.new_session(request, id)

        session.access_time = time()
        return session

    # get_session ()

    def maintain_session (self, request, session):
        """maintain_session(request : HTTPRequest, session : Session)

        Maintain session information.  This method should be called after
        servicing a HTTP request, just before the response is returned.  If
        a session contains information it is saved and a cookie dropped on
        the client.  If not, it the session is discarded.  It the client has
        a cookie it is deleted.
        """
        if self.sessions.has_key(session.id):
            # session is already saved
            if session.is_empty():
                # no use keeping it around
                del self.sessions[session.id]
                self.revoke_session_cookie(request)
        else:
            # session is not saved
            if not session.is_empty():
                # must save it in the dictionary of sessions and drop
                # a cookie on the client so that it can be found again
                self.sessions[session.id] = session
                self.set_session_cookie(request, session.id)


    def set_session_cookie (self, request, session_id):
        """Ensure that a session cookie with value 'session_id' will be
        returned to the client via 'response', and store it in 'reqeust'
        so that the cookie is available to any further processing of
        this request.
        """
        assert self.sessions.has_key(session_id), \
               "invalid session ID %s" % `session_id`
        config = get_publisher().config
        request.response.set_cookie(config.cookie_name, session_id,
                                    domain = config.cookie_domain,
                                    path = config.cookie_path)


    def revoke_session_cookie (self, request):
        """Remove the session cookie with value 'session_id' from the
        remote user's session by resetting the value and maximum age in
        'response'.  Also remove the cookie from 'request' so that
        further processing of this request does not see the cookie's
        revoked value.
        """
        config = get_publisher().config
        response = request.response
        response.set_cookie(config.cookie_name, "",
                            domain = config.cookie_domain,
                            path = config.cookie_path,
                            max_age = 0)
        if request.cookies.has_key(config.cookie_name):
            del request.cookies[config.cookie_name]


    def has_session (self, request, must_exist=0):
        """Return true if the current request already has a cookie
        identifying a session object.  If 'must_exist' is true, the
        cookie must correspond to a currently existing session;
        otherwise (the default), we just check for the existence of the
        session cookie and don't inspect its content at all.
        """
        config = get_publisher().config
        id = request.cookies.get(config.cookie_name)
        if id is None:
            return 0
        if must_exist:
            return self.sessions.has_key(id)
        else:
            return 1


    def get_app_state (self, request, app_id):
        """Fetch or create an application state object for the
        application identified by 'app_id' and the current session.
        """
        session = self.get_session(request)
        return session.get_app_state(app_id)


    def del_sessions (self, ip=None, age=None):
        """Delete a subset of open sessions selected according to the
        keyword arguments.  Default values select (and delete) no
        sessions.  Sessions are selected by combining non-None keyword
        args with "and": eg. if 'ip' and 'age' are non-None then any
        session matching *both* the supplied IP address and age (hours
        since session creation) is deleted.
        
        'ip' specifies an IP address (dotted-quad string); it must match
        the IP address of a session exactly.  'age' is the session's age
        (since creation) in hours; any session older than 'age' will
        pass the age test. Again, sessions must pass all specified tests
        (non-None arguments) to be selected and deleted.
        """
        goners = []
        for id in self.sessions.keys():
            (s_ip, s_date, s_time, rand) = string.split (id, ':')
            if ip is not None and ip != s_ip:
                continue
            if age is not None:
                datetime = strptime ("%s:%s" % (s_date, s_time), TIME_FORMAT)
                datetime = mktime (datetime)
            if (time() - datetime) <= float (age):
                continue

            # passed all the tests: we'll delete this sucker
            goners.append (id)

        return goners

    # del_session

# SessionManager


class Session:
    """Holds session information that crosses application boundaries,
    and maintains a registry of application state objects where
    individual applications keep their particular session-state
    information.

    Instance attributes:
      id : string
        the session ID (generated by SessionManager and used as the
        value of the session cookie)
      actual_user : any
        an object to identify the human being on the other end of the
        line.  It's up to you whether to store just a string in 'user',
        or some more complex data structure or object.
      user : any
        the user object used for access control checks.  Usually .user ==
        .actual_user, but for a situation where a user can act on
        behalf of another, you would leave .actual_user alone and
        set .user to another user object.
      creation_time : float
      access_time : float
        two ways of keeping track of the "age" of the whole session
        (without knowing anything about the age of individual
        application state objects).  Note that 'access_time' is
        maintained by the SessionManager that owns this session.
      app_state : {string : ApplicationState}
        maps application IDs to application state objects (which should
        be instances of ApplicationState subclasses)
      app_creation_time : {string : float}
      app_access_time : {string : float}
        keep track of the creation and last-access time for each
        application state object in 'app_state'

    Feel free to access 'id', 'user', 'creation_time', and 'access_time'
    directly.  'id' and the '_time' attributes must never change;
    if 'user' or 'actual_user' need to change, use 'set_user()' or
    'set_actual_user()'.  All other attributes are private: do not
    access or modify them. 
    """


    # 'app_state_class' - registry of application state classes (so
    # we know how to create the objects that go in the per-instance
    # application state object registry)
    app_state_class = {}

    MAX_FORM_TOKENS = 16 # maximum number of outstanding form tokens

    def __init__ (self, request, id):

        self.id = id
        self.user = self.actual_user = None
        self.creation_time = self.access_time = time()
        self.form_tokens = [] # queue
        self.app_state = {}
        self.app_creation_time = {}
        self.app_access_time = {}

    def __repr__ (self):
        return "<%s at %x: %s>" % (self.__class__.__name__, id(self), self.id)

    def __str__ (self):
        if self.actual_user is None:
            return "session %s (no user)" % self.id
        elif self.user == self.actual_user:
            return "session %s (user %s)" % (self.id, self.user)
        else:
            return "session %s (user %s working as %s)" % (self.id,
                                                           self.actual_user,
                                                           self.user)
    

    def is_empty (self):
        """Return true if this session contains no information that must be
        saved."""
        if self.actual_user or self.user or self.form_tokens:
            return 0
        for app in self.app_state.values():
            if not app.is_empty():
                return 0
        else:
            return 1

    def dump (self, file=None, header=1, deep=1):
        time_fmt = "%Y-%m-%d %H:%M:%S"
        ctime = strftime(time_fmt, localtime(self.creation_time))
        atime = strftime(time_fmt, localtime(self.access_time))

        if header:
            file.write('session %s:' % self.id)
        file.write('  user %s' % self.user)
        file.write('  created %s, last accessed %s' % (ctime, atime))
        file.write('  form_tokens: %s\n' % self.form_tokens)

        if deep:
            apps = self.app_state.keys()
            if apps:
                file.write('  application state objects:')
                apps.sort()
                for app in apps:
                    ctime = self.app_creation_time.get(app)
                    if ctime is not None:
                        ctime = strftime(time_fmt, localtime(ctime))
                    atime = self.app_access_time.get(app)
                    if atime is not None:
                        atime = strftime(time_fmt, localtime(atime))
                    file.write('%s: created %s, last accessed %s' %
                                  (app, ctime, atime))
                    app_state = self.app_state[app]
                    try:
                        st_lines = app_state.dump()
                    except (AttributeError, NotImplementedError):
                        file.write(app_state._dump())
                    else:
                        file.write(st_lines)
            else:
                file.write('no application state objects')
    
    # dump()


    # -- Hooks into the Quixote main loop ------------------------------

    def start_request (self, request):
        """Called near the beginning of each request: after the HTTPRequest
        object has been built and this Session object has been fetched or
        built, but before we traverse the URL or call the callable object
        found by URL traversal.
        """
        if self.user:
            request.environ['REMOTE_USER'] = str(self.user)

    def finish_request (self, request):
        """Called near the end of each request: after a callable object
        has been found and (successfully) called.  Not called if there
        were any errors processing the request.
        """
        pass


     # -- Simple accessors and modifiers --------------------------------

    def set_actual_user (self, user):
        self.user = self.actual_user = user

    def set_user (self, user):
        self.user = user

    def get_creation_age (self):
        """Return the number of seconds since session was created."""
        return time() - self.creation_time

    def get_access_age (self):
        """Return the number of seconds since session was last accessed."""
        return time() - self.access_time


    # -- More involved stuff -------------------------------------------

    def get_app_state (self, app_id):

        state = self.app_state.get(app_id)
        if state is None:
            klass = self.app_state_class.get(app_id)
            if klass is None:
                raise RuntimeError, \
                      "no application state class has been registered " + \
                      "for application '%s'" % app_id

            state = klass ()
            # XXX restore this check (using mems.lib.base.isinstance?)!
            if 0 and not isinstance (state, ApplicationState):
                raise RuntimeError, \
                      ("class %s (for application '%s') " +
                       "not a subclass of ApplicationState") % (klass, app_id)

            self.app_state[app_id] = state
            self.app_creation_time[app_id] = \
                self.app_access_time[app_id] = time()

        else:
            self.app_access_time[app_id] = time()

        return state

    # get_app_state ()


    def create_form_token (self):
        """create_form_token() -> string

        Create a new form token and add it to a queue of outstanding form
        tokens for this session.  A maximum of MAX_FORM_TOKENS are saved.
        The new token is returned.
        """
        token = "%016X" % randlong(8)
        self.form_tokens.append(token)
        extra = len(self.form_tokens) - self.MAX_FORM_TOKENS
        if extra > 0:
            del self.form_tokens[:extra]
        return token

    def has_form_token (self, token):
        """has_form_token(token : string) -> boolean

        Return true if 'token' is in the queue of outstanding tokens.
        """
        return token in self.form_tokens
        
    def remove_form_token (self, token):
        """remove_form_token(token : string)

        Remove 'token' from the queue of outstanding tokens.
        """
        self.form_tokens.remove(token)

# Session


class ApplicationState:
    """Defines the interface that application state classes must
    implement.

    Instance attributes: none
    """
    
    def __init__ (self):
        """Constructor takes no arguments, and just initializes all
        application session state structures.
        """

    def is_empty (self):
        """Return true if the ApplicationState is empty (i.e. it has no
        information that must be saved).  Subclasses should override this.
        Sessions that contain empty application states may be dropped in
        order to save memory."""
        return 0 # unless a smart subclass says otherwise

    def cleanup (self):
        """Destroy all application state data in memory without regard
        to persistence: the goal here is to free up memory, not to
        commit session data anywhere.
        """

    def dump (self, indent='', header=1, deep=1):
        """Dump all relevant and interesting information in this
        application state object.  If 'header' is true (the default),
        the first line of the dump will identify the object, and the
        rest of the dump will be indented deeper than the header line.
        Otherwise, the initial line is skipped and the whole dump is
        indented by 'indent'.  If 'deep' is true, we'll dump contained
        objects that might be bulky or non-essential; otherwise, we'll
        skip dumping them.
        """
        
# ApplicationState


def register_app_state_class (klass):
    try:
        app_id = klass.app_id
    except AttributeError:
        raise RuntimeError, \
              "application state class '%s' must define 'app_id' attribute" % \
              klass.__name__
    Session.app_state_class[app_id] = klass


#
# Pair of functions for getting and setting the singleton SessionManager
# object.
# 

_session_mgr = None

def get_session_manager ():
    global _session_mgr
    return _session_mgr

def set_session_manager (manager):
    global _session_mgr
    if _session_mgr is not None:
        raise RuntimeError, "session manager already set"
    _session_mgr = manager


    
# These are the functions commonly used by callers.
def get_session (request):
    return get_session_manager().get_session(request)

def has_session (request=None, must_exist=0):
    return get_session_manager().has_session(request, must_exist)

def get_app_state (request, app_id):
    if app_id is None:
        raise RuntimeError, "app_id must be supplied"
    return get_session_manager().get_app_state(request, app_id)



"""quixote.http_request

Provides the HTTPRequest class and related code for parsing HTTP
requests, such as the FileUpload class.

Derived from Zope's HTTPRequest module.
"""

##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################

__revision__ = "$Id: http_request.py,v 1.32 2001/10/09 19:40:42 gward Exp $"

import re
from cgi import FieldStorage
from urllib import quote, unquote
from types import ListType, StringType

from quixote.http_response import HTTPResponse


# Various regexes for parsing specific bits of HTTP, all from RFC 2616.

# These are needed by 'get_encoding()', to parse the "Accept-Encoding"
# header.  LWS is linear whitespace; the latter two assume that LWS
# has been removed.
_http_lws_re = re.compile("(\r\n)?[ \t]+")
_http_list_re = re.compile(r",+")
_http_encoding_re = re.compile(r"([^;]+)(;q=([\d.]+))?$")

# These are needed by 'guess_browser_version()', for parsing the
# "User-Agent" header.
#   token = 1*<any CHAR except CTLs or separators>
#   CHAR = any 7-bit US ASCII character (0-127)
#   separators are  ( ) < > @ , ; : \ " / [ ] ? = { }
#
# The comment RE is a simplification and not to spec -- it doesn't
# allow for backslash-quoting within comments or nested comments!
#
# The user_agent RE is also a simplification; it only looks for
# one "product", possibly followed by a comment.
_http_token_pat = r'[^\x00-\x20\(\)\<\>\@\,\;\:\\\"\/\[\]\?\=\{\}\x7F-\xFF]+'
_http_product_pat = r'(%s)(?:/(%s))?' % (_http_token_pat, _http_token_pat)
_http_product_re = re.compile(_http_product_pat)
_http_comment_pat = r'\(([^\(\)]*)\)'
_http_useragent_re = re.compile(r'%s(?:\s+%s)?'
                                % (_http_product_pat, _http_comment_pat))
_comment_delim_re = re.compile(r';\s*')



class HTTPRequest:
    """
    Model a single HTTP request and all associated data.
    
    This object provides access to request data.  This includes, the
    input headers, form data, server data, and cookies.

    The request object is a mapping object that represents a
    collection of variable to value mappings.  In addition, variables
    are divided into four categories:

      - Environment variables

        These variables include input headers, server data, and other
        request-related data.  The variable names are as specified in
        the CGI specification (although we're at the mercy of the
        web server to conform to that spec).

      - Form data

        These are data extracted from either a URL-encoded query
        string or body, if present.

      - Cookies

        These are the cookie data, if present.

    The form attribute of a request is actually a Field Storage
    object.  When file uploads are used, this provides a richer and
    more complex interface than is provided by accessing form data as
    items of the request.  See the FieldStorage class documentation
    for more details.
    """

    def __init__ (self, stdin, environ):
        self._file = None
        self.stdin = stdin
        self.environ = environ
        self.form = {}
        self.session = None
        self.response = HTTPResponse()

        # The strange treatment of SERVER_PORT_SECURE is because IIS
        # sets this environment variable to "0" for non-SSL requests
        # (most web servers -- well, Apache at least -- simply don't set
        # it in that case).
        if (environ.get('HTTPS', "off").lower() == "on" or
            environ.get('SERVER_PORT_SECURE', '0') != '0'):
            self.scheme = "https"
        else:
            self.scheme = "http"

        k = self.environ.get('HTTP_COOKIE','')
        if k:
            self.cookies = parse_cookie(k)
        else:
            self.cookies = {}

        # IIS breaks PATH_INFO because it leaves in the path to
        # the script, so SCRIPT_NAME is "/cgi-bin/q.py" and PATH_INFO
        # is "/cgi-bin/q.py/foo/bar".  The following code fixes 
        # PATH_INFO to the expected value "/foo/bar".
        web_server = environ.get('SERVER_SOFTWARE', 'unknown')
        if (web_server.find('Microsoft-IIS') != -1):
            script = environ['SCRIPT_NAME']
            path = environ['PATH_INFO']
            if path.startswith(script):
                path = path[len(script):]
                self.environ['PATH_INFO'] = path
                
    def process_inputs (self):
        """Process request inputs.
        """
        method = self.environ.get('REQUEST_METHOD','GET')
        
        if method != 'GET':
            fp = self.stdin
        else:
            fp = None

        fs = FieldStorage(fp=fp, environ=self.environ, keep_blank_values=1)
        if not hasattr(fs,'list') or fs.list is None:
            self._file = fs.file
        else:
            for item in fs.list:
                key = item.name
                if (hasattr(item,'file') and hasattr(item,'filename')
                    and hasattr(item,'headers')):
                    if (item.file and
                        (item.filename is not None
                         # RFC 1867 says that all fields get a content-type.
                         # or 'content-type' in map(lower, item.headers.keys())
                         )):
                        item = FileUpload(item)
                    else:
                        item = item.value

                #Insert in dictionary
                if self.form.has_key(key):
                    # it is not a record or list of records
                    found = self.form[key]
                    if type(found) is ListType:
                        found.append(item)
                    else:
                        found = [found,item]
                        self.form[key] = found
                else:
                    self.form[key] = item
                                
    def get_header (self, name, default=None):
        """get_header(name : string, default : string = None) -> string

        Return the named HTTP header, or an optional default argument
        (or None) if the header is not found.  Note that both original
        and CGI-ified header names are recognized, e.g. 'Content-Type',
        'CONTENT_TYPE' and 'HTTP_CONTENT_TYPE' should all return the
        Content-Type header, if available.
        """
        environ = self.environ
        name = ("_".join(name.split("-"))).upper()
        val = environ.get(name)
        if val is not None:
            return val
        if name[:5] != 'HTTP_':
            name = 'HTTP_%s' % name
        return environ.get(name, default)

    def get_cookie (self, cookie_name, default=None):
        return self.cookies.get(cookie_name, default)

    def get_form_var (self, var_name, default=None):
        return self.form.get(var_name, default)


    # The following environment variables are useful for reconstructing
    # the original URL, all of which are specified by CGI 1.1:
    #
    #   SERVER_NAME            "www.example.com"
    #   SCRIPT_NAME            "/q"
    #   PATH_INFO              "/debug/dump_sessions"
    #   QUERY_STRING           "session_id=10.27.8.40...."

    def get_server (self):
        """get_server() -> string

        Return the server name with an optional port number, eg.
        "www.example.com" or "foo.bar.com:8000".
        """
        server_name = self.environ["SERVER_NAME"].strip()
        server_port = self.environ.get("SERVER_PORT")

        if (not server_port or
            (self.scheme == "http" and server_port == "80") or
            (self.scheme == "https" and server_port == "443")):
            return server_name
        else:
            return server_name + ":" + server_port

    def get_path (self, n=0):
        """get_path(n : int = 0) -> string

        Return the path of the current request, chopping off 'n' path
        components from the right.  Eg. if the path is "/bar/baz/qux",
        n=0 would return "/bar/baz/qux" and n=2 would return "/bar".
        Note that the query string, if any, is not included.

        A path with a trailing slash should just be considered as having
        an empty last component.  Eg. if the path is "/bar/baz/", then:
          get_path(0) == "/bar/baz/"
          get_path(1) == "/bar/baz"
          get_path(2) == "/bar"

        Raises ValueError if n is too big for the current path.
        """
        path = "%(SCRIPT_NAME)s%(PATH_INFO)s" % self.environ
        if n == 0:
            return path
        else:
            path_comps = path.split('/')
            if n > len(path_comps)-1:
                raise ValueError, "n=%d too big for path '%s'" % (n, path)
            return '/'.join(path_comps[:-n])

    def get_url (self, n=0):
        """get_url(n : int = 0) -> string

        Return the URL of the current request, chopping off 'n' path
        components from the right.  Eg. if the URL is
        "http://foo.com/bar/baz/qux", n=2 would return
        "http://foo.com/bar".  Does not include the query string (if
        any).
        """
        return "%s://%s%s" % (self.scheme, self.get_server(), self.get_path(n))

    def get_environ (self, key, default=None):
        """get_environ(key : string) -> string

        Fetch a CGI environment variable from the request environment.
        See http://hoohoo.ncsa.uiuc.edu/cgi/env.html
        for the variables specified by the CGI standard.
        """
        return self.environ.get(key, default)

    def get_encoding (self, encodings):
        """get_encoding(encodings : [string]) -> string
           
        Parse the "Accept-encoding" header. 'encodings' is a list of
        encodings supported by the server sorted in order of preference.
        The return value is one of 'encodings' or None if the client
        does not accept any of the encodings.
        """
        found_encodings = {}
        accept_encoding = self.get_header("accept-encoding") or ""
        # remove all linear whitespace
        accept_encoding = _http_lws_re.sub("", accept_encoding)
        for coding in _http_list_re.split(accept_encoding):
            m = _http_encoding_re.match(coding)
            if m:
                encoding = m.group(1).lower()
                q = m.group(3) or 1
                try:
                    q = float(q)
                except ValueError:
                    continue
                if encoding == "*":
                    continue # stupid, ignore it
                if q > 0:
                    found_encodings[encoding] = 1
        if found_encodings:
            for encoding in encodings:
                if found_encodings.has_key(encoding):
                    return encoding
        return None

    def dump_html (self):
        result="<h3>form</h3><table>"
        row='<tr valign="top" align="left"><th>%s</th><td>%s</td></tr>'
        for k,v in self.form.items():
            result=result + row % (html_quote(k), html_quote(v))
        result=result+"</table><h3>cookies</h3><table>"
        for k,v in self.cookies.items():
            result=result + row % (html_quote(k), html_quote(v))
    
        result=result+"</table><h3>environ</h3><table>"
        for k,v in self.environ.items():
            result=result + row % (html_quote(k), html_quote(v))
        return result+"</table>"

    def dump (self):
        result=""
        row='%-15s %s\n'
        L = self.form.items() ; L.sort()
        for k,v in L:
            result=result + row % (k,v)
        result=result+"\nCookies:\n"
        L = self.cookies.items() ; L.sort()
        for k,v in L:
            result=result + row % (k,v)

        result=result+"\nEnvironment:\n"
        L = self.environ.items() ; L.sort()
        for k,v in L:
            result=result + row % (k,v)
        return result

    def guess_browser_version (self):
        """guess_browser_version() -> (name : string, version : string)

        Examine the User-agent request header to try to figure out what
        the current browser is.  Returns a (name, version) tuple where
        each element is a string or None if we couldn't parse the
        User-agent header.  Handles Microsoft's little joke of
        pretending to be Mozilla, eg. if the "User-Agent" header is
        "Mozilla/5.0 (compatible; MSIE 5.5)" returns ("MSIE", "5.5").
        Konqueror does the same thing, and it's handled the same way.
        """
        ua = self.get_header('user-agent')
        if ua is None:
            return (None, None)

        # The syntax for "User-Agent" in RFC 2616 is fairly simple:
        #   User-Agent = "User-Agent" ":" 1*( product | comment )
        #   product    = token ["/" product-version]
        #   product-version = token
        #   token      = 1*<any CHAR except CTLs or separators>
        #
        # The _http_useragent_re handles the most-commonly-used
        # subset of this syntax, namely
        #   User-Agent = "User-Agent" ":" product 1*SP [comment]
        # ie. one product string followed by an optional comment;
        # anything after that first comment is ignored.  This should be
        # enough to distinguish Mozilla/Netscape, MSIE, Opera, and
        # Konqueror.

        ua_match = _http_useragent_re.match(ua)
        if not ua_match:
            import sys
            sys.stderr.write("couldn't parse User-Agent header: %r\n" % ua)
            return (None, None)

        (name, version, comment) = ua_match.groups()
        if comment:
            comment_chunks = _comment_delim_re.split(comment)
        else:
            comment_chunks = []

        if ("compatible" in comment_chunks and
            len(comment_chunks) > 1 and comment_chunks[1]):
            # A-ha!  Someone is kidding around, pretending to be what
            # they are not.  Most likely MSIE masquerading as Mozilla,
            # but lots of other clients (eg. Konqueror) do the same.
            real_ua = comment_chunks[1]
            if "/" in real_ua:
                (name, version) = real_ua.split("/", 1)
            else:
                if real_ua.startswith("MSIE"):
                    (name, version) = real_ua.split(" ", 1)
                else:
                    name = real_ua
                    version = None
            return (name, version)

        # Either nobody is pulling our leg, or we didn't find anything
        # that looks vaguely like a user agent in the comment.  So use
        # what we found outside the comment, ie. what the spec says we
        # should use (sigh).
        return (name, version)

    # guess_browser_version ()
    
    def redirect (self, location, permanent=0):
        """redirect(location : string, permanent : boolean = false)
           -> string

        Create a redirection response.  If the location does not
        specify the server then it is automatically added.  The return
        value is an HTML document indicating the new URL (useful if the
        client browser does not honor the redirect).
        """
        if location.find("://") == -1:
            location = "%s://%s%s" % (self.scheme, self.get_server(), location)
        return self.response.redirect(location, permanent)


class FileUpload:
    """
    Used to represent file-uploaded data.  Can be used just like files.

    In addition, they have a 'headers' attribute that is a dictionary
    containing the file-upload headers, and a 'filename' attribute
    containing the name of the uploaded file.
    """

    def __init__ (self, fields):
        """FileUpload(fields : FieldStorage)
        """

        file = fields.file
        if hasattr(file, '__methods__'):
            methods = file.__methods__
        else:
            methods = ['close', 'fileno', 'flush', 'isatty',
                       'read', 'readline', 'readlines', 'seek',
                       'tell', 'truncate', 'write', 'writelines']

        d = self.__dict__
        for m in methods:
            if hasattr(file,m):
                d[m] = getattr(file,m)

        self.headers = fields.headers
        self.filename = fields.filename
    

_qparm_re = re.compile(r'([\0- ]*'
                       r'([^\0- ;,=\"]+)="([^"]*)"'
                       r'([\0- ]*[;,])?[\0- ]*)')
_parm_re = re.compile(r'([\0- ]*'
                      r'([^\0- ;,="]+)=([^\0- ;,"]*)'
                      r'([\0- ]*[;,])?[\0- ]*)')

def parse_cookie (text):
    result = {}

    pos = 0
    while 1:
        mq = _qparm_re.match(text, pos)
        m = _parm_re.match(text, pos)
        if mq is not None:
            # Match quoted correct cookies
            name = mq.group(2)
            value = mq.group(3)
            pos = mq.end()
        elif m is not None:
            # Match evil MSIE cookies ;)
            name = m.group(2)
            value = m.group(3)
            pos = m.end()
        else:
            # this may be an invalid cookie.
            # We'll simply bail without raising an error
            # if the cookie is invalid.
            return result

        if not result.has_key(name):
            result[name] = value

    return result



import string
from types import *

class IndentedFile:
    """File-like object that adds indentation to each output line.
    Supports all file methods except 'read()', 'readline()',
    'readlines()', and 'readinto()'."""

    def __init__ (self, file, indent=2, base_indent=None):
        """Initialize a new IndentedFile.  'file' must be a file or
        file-like object opened for writing; 'indent', if given, is the
        string to prepend to each line *or* the number of spaces to
        prepend."""

        self.file = file
        if type (indent) is IntType:
            self.indent = ' ' * indent
        elif type (indent) is StringType:
            self.indent = indent
        else:
            raise TypeError, "'indent' must be a string or an integer"

        if base_indent is None:
            self.base_indent = indent
        else:
            self.base_indent = base_indent
        self.at_start = 1               # at start of a line?

        for method_name in ('close', 'tell', 'seek', 'isatty',
                            'flush', 'truncate', 'fileno'):

            # Don't get uptight if we can't find a method on the
            # supposed file-like object -- it might be a user-defined
            # file-like object, or this might be JPython, where even
            # standard file objects don't provide 'isatty()' or
            # 'truncate()'.
            try:
                setattr (self, method_name,
                         getattr (self.file, method_name))
            except AttributeError:
                pass


    def write (self, str):
        lines = string.split (str, '\n')

        if len (lines) == 1:            # no newlines
            if self.at_start:
                self.file.write (self.indent)
            self.file.write (lines[0])
            self.at_start = 0           # next 'write' continues this line

        else:

            # write the first line (which might be a continuation of
            # a previous line)
            if self.at_start:
                self.file.write (self.indent)
            self.file.write (lines[0] + '\n')

            # now write all the complete lines in 'str', ie. everything
            # up to the final newline
            for l in lines[1:-1]:
                self.file.write (self.indent + l + '\n')
            self.at_start = 1

            # and write any text that follows the final newline
            if lines[-1]:
                self.file.write (self.indent + lines[-1])
                self.at_start = 0


    def writelines (self, lines):
        for l in lines:
            self.write (l)


    def indent_more (self, indent=None):

        if indent is None:
            self.indent_more (self.base_indent)
            return

        if type (indent) is IntType:
            self.indent = self.indent + (' ' * indent)
        elif type (indent) is StringType:
            self.indent = self.indent + indent
        else:
            raise TypeError, "'indent' must be a string or an integer"

    def indent_less (self, indent=None):

        if indent is None:
            self.indent_less (self.base_indent)
            return

        if type (indent) is IntType:
            self.indent = self.indent[0:-indent]
        elif type (indent) is StringType:
            if self.indent[-len(indent):] == indent:
                self.indent = self.indent[0:-len(indent)]
            else:
                raise ValueError, \
                      "indentation string '%s' does not end in '%s'" % \
                      (self.indent, indent)
        else:
            raise TypeError, "'indent' must be a string or an integer"


if __name__ == "__main__":
    import os

    f = open ("foo", "w")
    indf = IndentedFile (f)
    indf.write ("Line 1\n")
    indf.write ("Line 2\nLine 3")
    indf.write ("...continued\n")
    indf.indent_more()
    indf.writelines (["Line 4\n", "Line 5", "...continued\n", "Line 6"])
    indf.indent_less()
    indf.write ("\n")
    indf.write ("Line 7") ; indf.write ("\n")
    indf.indent_less()
    indf.write ("Line 8\nLine 9\n")
    indf.close()

    expect = ["  Line 1",
              "  Line 2",
              "  Line 3...continued",
              "    Line 4",
              "    Line 5...continued",
              "    Line 6",
              "  Line 7",
              "Line 8",
              "Line 9",
             ]

    print "Expected output:"
    for line in expect:
        print line

    print "Actual output:"
    os.system ("cat foo")
    



#!/usr/bin/env python

# Setup script for Quixote

__revision__ = "$Id: setup.py,v 1.12 2001/10/10 18:01:41 gward Exp $"

from distutils.core import setup
from qx_distutils import qx_build_py

setup (name = "Quixote",
       version = "0.4.1",
       description = "A highly Pythonic Web application framework",
       author = "MEMS Exchange",
       author_email = "quixote@mems-exchange.org",
       url = "http://www.mems-exchange.org/software/python/quixote/",

       package_dir = {'quixote':''},
       packages = ['quixote',  'quixote.demo', 'quixote.form'],

       cmdclass = {'build_py': qx_build_py},
      )

 
